"""

Ansible METTA provisioner plugin

This package is a WIP and currently doesn't really do anything.

"""

import os
import logging
from typing import Any, Dict

import toml
import yaml

from configerus.loaded import Loaded, LOADED_KEY_ROOT

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures

from .ansiblecli_client import (
    METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID,
    METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKCLIENT_PLUGIN_ID,
)

logger = logging.getLogger("metta.contrib.provisioner:ansible")

METTA_ANSIBLE_ANSIBLECLIPLAYBOOK_PROVISIONER_PLUGIN_ID = "metta_ansiblecliplaybook_provisioner"
""" Ansible provisioner plugin id """

ANSIBLE_PROVISIONER_CONFIG_LABEL = "ansible"
""" config label loading the ansible config """

ANSIBLE_PROVISIONER_CONFIG_ENVS_KEY = "envs"
"""Configerus key for loading env variables to be used for clients."""
ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_KEY = "playbook.contents"
""" config key for the ansible playbook content """
ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_PATH_KEY = "playbook.path"
""" config key for the ansible playbook path to write to."""
ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_VARS_KEY = "vars.values"
""" config key for the ansible playbook vars."""
ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_VARS_PATH_KEY = "vars.path"
""" config key for the ansible playbook vars path to write to."""
ANSIBLE_PROVISIONER_CONFIG_ANSIBLECFG_KEY = "ansiblecfg.contents"
""" config key for the ansible cfg contents which will be written to file """
ANSIBLE_PROVISIONER_CONFIG_ANSIBLECFG_PATH_KEY = "ansiblecfg.path"
""" config key for the ansible cfg file path to write to."""
ANSIBLE_PROVISIONER_CONFIG_INVENTORY_KEY = "inventory.contents"
""" config key for the ansible  inventory, which will be passed to ansible """
ANSIBLE_PROVISIONER_CONFIG_INVENTORY_PATH_KEY = "inventory.path"
""" config key for the ansible inventory file path to write to."""


# pylint: disable=too-many-instance-attributes
class AnsiblePlaybookProvisionerPlugin:
    """Ansible provisioner plugin.

    Provisioner plugin that allows control of and interaction with a ansible
    cluster.

    ## Requirements

    1. this plugin uses subprocess to call a ansible binary, so you have to install
       ansible in the environment

    ## Usage

    @NOTE his plugin ist still very much a stub, and implements no real functionality.

    """

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        label: str = ANSIBLE_PROVISIONER_CONFIG_LABEL,
        base: Any = LOADED_KEY_ROOT,
    ):
        """Keep metta info and plugin config info.

        Parameters:
        -----------
        environment (Environment) : All metta plugins receive the environment
            object in which they were created.
        instance_id (str) : all metta plugins receive their own string identity.

        label (str) : Configerus load label for finding plugin config.
        base (str) : Configerus get base key in which the plugin should look for
            all config.

        """
        self._environment: Environment = environment
        """Environemnt in which this plugin exists."""
        self._instance_id: str = instance_id
        """Unique id for this plugin instance."""

        self._config_label: str = label
        """ configerus load label that should contain all of the config """
        self._config_base: str = base
        """ configerus get key that should contain all plugin config """

        self.fixtures: Fixtures = Fixtures()
        """Fixtures created by this plugin - typically various clients."""

        # In order to allow declarative interaction. Try to make an ansible client
        # for this plugin, but allow it to fail.
        try:
            self._update_config()
            self._make_clients()

        # no exception in init should block building the object
        # pylint: disable=broad-except
        except Exception:
            logger.debug("Inital ansible plugin build failed : %s", self._instance_id)

    # deep argument is an info() standard across plugins
    # pylint: disable=unused-argument
    def info(self, deep: bool = False):
        """Get info about a provisioner plugin."""
        # Loaded configerus config for the plugin. Ready for .get().
        plugin_config: Loaded = self._environment.config().load(self._config_label)

        return {
            "config": {
                "config_label": self._config_label,
                "config_base": self._config_base,
            },
            "ansible": {
                "ansiblecfg": plugin_config.get(
                    [self._config_base, ANSIBLE_PROVISIONER_CONFIG_ANSIBLECFG_KEY],
                    default="NONE",
                ),
                "inventory": plugin_config.get(
                    [self._config_base, ANSIBLE_PROVISIONER_CONFIG_INVENTORY_KEY],
                    default="MISSING",
                ),
                "playbook": plugin_config.get(
                    [self._config_base, ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_KEY],
                    default="MISSING",
                ),
                "vars": plugin_config.get(
                    [self._config_base, ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_VARS_KEY],
                    default="NONE",
                ),
                "envs": plugin_config.get(
                    [self._config_base, ANSIBLE_PROVISIONER_CONFIG_ENVS_KEY],
                    default="NONE",
                ),
            },
            "files": {
                "ansiblecfg": plugin_config.get(
                    [self._config_base, ANSIBLE_PROVISIONER_CONFIG_ANSIBLECFG_PATH_KEY],
                    default="NONE",
                ),
                "inventory": plugin_config.get(
                    [self._config_base, ANSIBLE_PROVISIONER_CONFIG_INVENTORY_PATH_KEY],
                    default="MISSING",
                ),
                "playbook": plugin_config.get(
                    [self._config_base, ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_PATH_KEY],
                    default="MISSING",
                ),
                "vars": plugin_config.get(
                    [self._config_base, ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_VARS_PATH_KEY],
                    default="NONE",
                ),
            },
        }

    def prepare(self):
        """Prepare the provisioner to apply resources.

        Initial Provisioner plugin is expected to be of very low cost until
        prepare() is executed.  At this point the plugin should load any config
        and perform any validations needed.
        The plugin should not create any resources but it is understood that
        there may be a cost of preparation.

        Provisioners are expected to load a lot of config to self-program.
        Because of this, they allow passing of a configerus label for .load()
        and a base for .get() in case there is an alterante config source
        desired.

        """

    def apply(self):
        """Bring a cluster to the configured state.

        Run an ansible playbook to execute functionality on a running cluster.

        """
        self._update_config()
        self._make_clients()

        # Client plugin which will do all the interaction for us.
        playbook_client = self.fixtures.get_plugin(
            plugin_id=METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKCLIENT_PLUGIN_ID
        )

        # Loaded configerus config for the plugin. Ready for .get().
        plugin_config: Loaded = self._environment.config().load(
            self._config_label, force_reload=True
        )

        playbook_path: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_PATH_KEY]
        )
        vars_path: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_VARS_PATH_KEY],
            default="",
        )
        envs: Dict[str, str] = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_ENVS_KEY],
            default={},
        )

        playbook_client.run_file(
            playbooksyml_path=playbook_path, extravars_path=vars_path, envs=envs
        )

    def destroy(self):
        """Remove all resources created for the cluster.

        Run an aansible palybook to remove any implemented changes

        @NOTE this plugin ist still very much a stub, and implements no real functionality.

        """
        self._rm_ansible_files()

    def _make_clients(self):
        """Create and assign an ansible client to this plugin."""
        # Loaded configerus config for the plugin. Ready for .get().
        plugin_config: Loaded = self._environment.config().load(self._config_label)

        ansiblecfg_path: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_ANSIBLECFG_PATH_KEY],
            default="",
        )
        inventory_path: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_INVENTORY_PATH_KEY]
        )

        inventory_contents: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_INVENTORY_KEY]
        )
        playbook_contents: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_KEY], default={}
        )

        # Don't create the object unless we think we hae enough config for it
        if not (inventory_path and inventory_contents and playbook_contents):
            raise RuntimeError("No inventory provided so we are not creating an ansible client.")

        instance_id = f"{self._instance_id}-{METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID}"
        fixture = self._environment.new_fixture(
            plugin_id=METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID,
            instance_id=instance_id,
            priority=70,
            arguments={
                "ansiblecfg_path": ansiblecfg_path,
                "inventory_path": inventory_path,
            },
            replace_existing=True,
        )
        self.fixtures.add(fixture, replace_existing=True)

        instance_id = f"{self._instance_id}-{METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKCLIENT_PLUGIN_ID}"
        fixture = self._environment.new_fixture(
            plugin_id=METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKCLIENT_PLUGIN_ID,
            instance_id=instance_id,
            priority=70,
            arguments={
                "ansiblecfg_path": ansiblecfg_path,
                "inventory_path": inventory_path,
            },
            replace_existing=True,
        )
        self.fixtures.add(fixture, replace_existing=True)

    def _update_config(self):
        """Update config and write the cfg and inventory files."""
        # refresh any loaded config
        plugin_config = self._environment.config().load(self._config_label, force_reload=True)

        # first the ansible cfg file
        ansiblecfg_contents: Dict[str, Any] = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_ANSIBLECFG_KEY], default={}
        )
        ansiblecfg_path: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_ANSIBLECFG_PATH_KEY],
            default="",
        )
        if ansiblecfg_contents:
            # ansible doesn't like how toml keeps some quotes - so we have to do this ourselves.
            ansiblecfg_contents = toml.dumps(ansiblecfg_contents)
            ansiblecfg_contents = ansiblecfg_contents.replace('"', "")
            os.makedirs(os.path.dirname(os.path.realpath(ansiblecfg_path)), exist_ok=True)
            with open(ansiblecfg_path, "w", encoding="utf8") as ansiblecfg_fileobject:
                ansiblecfg_fileobject.write(ansiblecfg_contents)

        else:
            if ansiblecfg_path and os.path.exists(ansiblecfg_path):
                os.remove(ansiblecfg_path)

        # second the inventory file
        inventory_contents: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_INVENTORY_KEY], default={}
        )
        inventory_path: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_INVENTORY_PATH_KEY],
            default="",
        )
        if inventory_contents:
            os.makedirs(os.path.dirname(os.path.realpath(inventory_path)), exist_ok=True)
            with open(inventory_path, "w", encoding="utf8") as inventory_fileobject:
                inventory_fileobject.write(inventory_contents)
        else:
            if inventory_path and os.path.exists(inventory_path):
                os.remove(inventory_path)

        # the playbook file
        playbook_contents: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_KEY], default={}
        )
        playbook_path: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_PATH_KEY],
            default="",
        )
        if playbook_contents:
            os.makedirs(os.path.dirname(os.path.realpath(playbook_path)), exist_ok=True)
            with open(playbook_path, "w", encoding="utf8") as playbook_fileobject:
                yaml.safe_dump(playbook_contents, playbook_fileobject)
        else:
            if playbook_path and os.path.exists(playbook_path):
                os.remove(playbook_path)

        # the playbook vars file
        vars_values: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_VARS_KEY], default={}
        )
        vars_path: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_VARS_PATH_KEY],
            default="",
        )
        if vars_values:
            os.makedirs(os.path.dirname(os.path.realpath(vars_path)), exist_ok=True)
            with open(vars_path, "w", encoding="utf8") as vars_fileobject:
                yaml.safe_dump(vars_values, vars_fileobject)
        else:
            if vars_path and os.path.exists(vars_path):
                os.remove(vars_path)

    def _rm_ansible_files(self):
        """Update config and write the cfg and inventory files."""
        logger.info("Ansible provisioner removing created files.")

        # Loaded configerus config for the plugin. Ready for .get().
        plugin_config: Loaded = self._environment.config().load(self._config_label)

        # first the ansible cfg file
        ansiblecfg_path: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_ANSIBLECFG_PATH_KEY],
            default="",
        )
        if ansiblecfg_path and os.path.exists(ansiblecfg_path):
            os.remove(ansiblecfg_path)

        # second the inventory file
        inventory_path: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_INVENTORY_PATH_KEY],
            default="",
        )
        if inventory_path and os.path.exists(inventory_path):
            os.remove(inventory_path)

        # the playbook file
        playbook_path: str = plugin_config.get(
            [self._config_base, ANSIBLE_PROVISIONER_CONFIG_PLAYBOOK_PATH_KEY],
            default="",
        )
        if playbook_path and os.path.exists(playbook_path):
            os.remove(playbook_path)



"""Callback we can use with the ansible client."""

from typing import Dict, Any, Iterator, Tuple
from enum import Enum

from ansible.plugins.callback import CallbackBase


class ResultStatus(Enum):
    """Result status string as an enum."""

    OK = "ok"
    UNREACHABLE = "unreachable"
    FAILED = "failed"


# Yes this is a struct.  No I am not sorry.
# pylint: disable=too-few-public-methods
class Result:
    """An ansible result from a results callback."""

    def __init__(self, status: ResultStatus, host: str, result: Dict[str, Any], *args, **kwargs):
        """Parametrize from a result Dict."""
        self.status: ResultStatus = status
        self.host: str = host
        self.result: Dict[str, Any] = result
        self.args: Tuple[Any] = args
        self.kwargs: Dict[str, Any] = kwargs


class ResultsCallback(CallbackBase):
    """A sample callback plugin used for collecting results as they come in."""

    def __init__(self, *args, **kwargs):
        """Initialize a results set."""
        super().__init__(*args, **kwargs)
        self._results: Dict[str, Result] = {}
        """Per host ResultStatus for an operation."""

    def __len__(self) -> int:
        """Return how many results we have.

        Returns:
        --------
        Integer number or results received.

        """
        return len(self._results)

    def __getitem__(self, host: str) -> Result:
        """Subscribe to a result using host name.

        Parameters:
        -----------
        hots (str) : Hostname.

        Returns:
        --------
        A Result object for the passed host name.

        Raises:
        -------
        Will raise a KeyError if the instance_id does not exist in the set.

        """
        return self._results[host]

    def __iter__(self) -> Iterator[Result]:
        """Create an iterator for the fixtures object.

        @TODO switch to just using a generator?

        Returns:
        --------
        An Iterator of Fixture objects

        """
        # Iterate across the to_list() set, as it is sorted.
        return iter(self._results.values())

    def __reversed__(self) -> Iterator[Result]:
        """Create a reversed iterator for the results.

        @TODO switch to just using a generator?

        Returns:
        --------
        An Iterator of Result objects

        """
        # Iterate across the to_list() set, as it is sorted.
        return reversed(self._results.values())

    # ANSIBLE CALLBACK FUNCTIONALITY
    #
    # The following methods are called by the ansible queue manager strategy object while the queue
    # is being processed.  Here we just try to catch information to produce discoverable results
    # in the object.
    #

    # need to access result _result in order to get info out of it.
    # pylint: disable=protected-access

    def v2_runner_on_unreachable(self, result):
        """Register and unreachable runner."""
        self._results[result._host.get_name()] = Result(
            status=ResultStatus.UNREACHABLE,
            host=result._host,
            result=result._result,
        )

    def v2_runner_on_ok(self, result, *args, **kwargs):
        """Register an OK runner."""
        self._results[result._host.get_name()] = Result(
            status=ResultStatus.OK, host=result._host, result=result._result, *args, **kwargs
        )

    def v2_runner_on_failed(self, result, *args, **kwargs):
        """Register a failed runner."""
        self._results[result._host.get_name()] = Result(
            status=ResultStatus.FAILED, host=result._host, result=result._result, *args, **kwargs
        )



"""

Metta CLI : Output commands.

Various commands that allow introspection of output plugins/fixtures and
their contents.

"""
import logging
from typing import List


from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.output import METTA_PLUGIN_INTERFACE_ROLE_OUTPUT

from .base import CliBase, cli_output

logger = logging.getLogger("metta.cli.output")


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class OutputCliPlugin(CliBase):
    """Fire command/group generator for output commands."""

    def fire(self):
        """Return a dict of commands."""
        return {"output": OutputGroup(self._environment)}


class OutputGroup:
    """Base Fire command group for output commands."""

    def __init__(self, environment: Environment):
        """Create CLI command group."""
        self._environment: Environment = environment

    # pylint: disable=too-many-arguments
    def _filter(
        self,
        plugin_id: str = "",
        instance_id: str = "",
        has_labels: List[str] = None,
    ):
        """Filter fixtures centrally."""
        return self._environment.fixtures().filter(
            plugin_id=plugin_id,
            instance_id=instance_id,
            has_labels=has_labels,
            interfaces=METTA_PLUGIN_INTERFACE_ROLE_OUTPUT,
        )

    # methods needs to be on object for cli registation
    # pylint: disable=no-self-use
    def plugins(
        self,
        has_label: str = "",
    ):
        """List registered plugins and interfaces."""
        plugins_info = {}

        # use a protected property just for introspection
        # pylint: disable=protected-access
        for registration in Factory._registry.values():
            if METTA_PLUGIN_INTERFACE_ROLE_OUTPUT not in registration.interfaces:
                continue
            if has_label and has_label not in registration.labels:
                continue

            plugins_info[registration.plugin_id] = {
                "plugin_id": registration.plugin_id,
                "interfaces": registration.interfaces,
                "labels_from_factory": registration.labels,
            }

        return cli_output(plugins_info)

    # this is what is needed to limit or filter fixture introspection
    # pylint: disable=too-many-arguments
    def info(
        self,
        deep: bool = False,
        children: bool = False,
        plugin_id: str = "",
        instance_id: str = "",
        has_label: str = "",
    ):
        """Return Info for fixtures."""
        fixture_info_list = []
        for fixture in self._filter(
            plugin_id=plugin_id,
            instance_id=instance_id,
            has_labels=[has_label] if has_label else [],
        ):
            fixture_info_list.append(fixture.info(deep=deep, children=children))

        return cli_output(fixture_info_list)

    # this is what is needed to limit or filter fixture introspection
    # pylint: disable=too-many-arguments
    def list(
        self,
        plugin_id: str = "",
        instance_id: str = "",
        has_label: str = "",
    ):
        """Return Info for fixtures."""
        fixture_info_list = []
        for fixture in self._filter(
            plugin_id=plugin_id,
            instance_id=instance_id,
            has_labels=[has_label] if has_label else [],
        ):
            fixture_info_list.append(
                {
                    "plugin_id": fixture.plugin_id,
                    "instance_id": fixture.instance_id,
                    "interfaces": fixture.interfaces,
                    "labels": fixture.labels,
                }
            )

        return cli_output(fixture_info_list)



"""

Metta - A testing system management framework.

In this module, the root package, are found a number of tool plugin constructors
that can be used to get instances of the plugins based on either passed in data
or configerus config.

The contstruction.py module does the heavy lifting of building plugins and
PluginInstances sets.

"""
import os
from typing import List, Callable, Dict, Any
import logging

from configerus.config import Config

from .plugin import Factory
from .config import default_config

# this also registers the core bootstrap plugins
from .bootstrap import METTA_BOOTSTRAPPER_PROJECT_PLUGIN_ID as DEFAULT_BOOTSTRAPPER
from .fixture import Fixture

# This also registers the core environment plugins
from .environment import (
    Environment,
    METTA_BUILDER_ENVIRONMENT_PLUGIN_ID as DEFAULT_ENVIRONMENT,
    METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENT,
)
from .globals import global_fixtures

logger = logging.getLogger("metta")

""" METTA and Configerus bootstrapping """

FIXED_CONFIGERUS_BOOSTRAPS = ["get", "env", "jsonschema", "dict", "files"]
""" configerus bootstraps that we will use on config objects. We use functionality
    in the core which heavily depends on this config, so it is necessary """


def discover(
    boostrapper_plugin_id: str = DEFAULT_BOOTSTRAPPER,
    config: Config = None,
    labels: Dict[str, str] = None,
    arguments: Dict[str, Any] = None,
) -> Fixture:
    """Ask Metta to discover and generate everything from config.

    Do you prefer to just write some config, and then have Metta do all of the
    preparation for you?

    You can optionally provide some of the components, but if you don't Metta
    will create what it needs on the fly.

    Parameters:
    -----------

    bootrapper (str) : Bootstrapper plugin id for the function which should be
        used to bootstrap this metta session.

    config (Config) : optionally provide a configerus Config object in if you
        want to provide prepopulated config.
        This method will make one if needed.

    """
    if config is None:
        config = default_config(FIXED_CONFIGERUS_BOOSTRAPS)

    instance_id = "bootstrapper"
    args: List[Any] = [config, instance_id]
    kwargs: Dict[str, Any] = arguments if arguments is not None else {}
    plugin_instance = Factory.create(boostrapper_plugin_id, instance_id, *args, **kwargs)

    # Build a fixture from the plugin_instance and add it to the fixtures
    # set for the environment, then return the fixture.
    fixture = global_fixtures.add(
        fixture=Fixture.from_instance(
            plugin_instance, priority=95, labels=labels if labels is not None else {}
        ),
        replace_existing=True,
    )
    return fixture


# pylint: disable=too-many-arguments
# This is what it takes to build a plugin.
def new_environment(
    name: str,
    config: Config = None,
    environment_plugin_id: str = DEFAULT_ENVIRONMENT,
    priority: int = 70,
    replace_existing: bool = False,
    labels: Dict[str, str] = None,
    arguments: Dict[str, Any] = None,
) -> Fixture:
    """Environment only way to use Metta.

    Not only do you not want to know about Spaces, but you actually don't want
    all of that, and can manage your own business.  Use this and get just an
    Environment object.

    You can optionally provide some of the components, but if you don't Metta
    will create what it needs on the fly.

    """
    if config is None:
        config = default_config(FIXED_CONFIGERUS_BOOSTRAPS)

    instance_id = name
    args: List[Any] = [config, instance_id]
    kwargs: Dict[str, Any] = arguments if arguments is not None else {}
    plugin_instance = Factory.create(environment_plugin_id, instance_id, *args, **kwargs)

    # Build a fixture from the plugin_instance and add it to the fixtures
    # set for the environment, then return the fixture.
    fixture = global_fixtures.add(
        fixture=Fixture.from_instance(
            plugin_instance, priority=priority, labels=labels if labels is not None else {}
        ),
        replace_existing=replace_existing,
    )
    return fixture


def environment_names() -> List[str]:
    """List all of the environment names that have been created."""
    return list(
        fixture.instance_id
        for fixture in global_fixtures.filter(interfaces=[METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENT])
    )


def get_environment(name: str = None) -> Environment:
    """Retrieve an environment.

    If no environment name is passed then the highest priority environment is returned.
    """
    return global_fixtures.get_plugin(
        instance_id=name, interfaces=[METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENT]
    )


def new_plugin(plugin_id: str, instance_id: str, *args, **kwargs):
    """Generate a plugin with your own values.

    Are you so on top of it that you just want to use the plugin registration
    and creation part of Metta.  With this method you are welcome to take
    advantage of the plugin system, but are expected to handle everything
    yourself.

    It is up to the consumer to make sure that any desired plugins are already
    registered, which can usually be done by simply importing the python modules
    that container the decorated factory functions.

    """
    priority = 75
    labels: Dict[str, str] = {"parent": "global"}
    plugin_instance = Factory.create(plugin_id, instance_id, *args, **kwargs)
    fixture = global_fixtures.add(
        fixture=Fixture.from_instance(plugin_instance, priority=priority, labels=labels),
        replace_existing=True,
    )
    return fixture.plugin



"""

Metta kubernetes CLI plugin.

Provides functionality to manually interact with the various kubernetes plugins.

"""
import logging
from typing import Dict, Any
from subprocess import CalledProcessError

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures
from mirantis.testing.metta_cli.base import CliBase, cli_output

from .kubeapi_client import METTA_PLUGIN_ID_KUBERNETES_CLIENT
from .helm_workload import METTA_PLUGIN_ID_KUBERNETES_HELM_WORKLOAD
from .yaml_workload import METTA_PLUGIN_ID_KUBERNETES_YAML_WORKLOAD
from .deployment_workload import METTA_PLUGIN_ID_KUBERNETES_DEPLOYMENT_WORKLOAD

logger = logging.getLogger("metta.cli.kubernetes")

METTA_PLUGIN_ID_KUBERNETES_CLI = "metta_kubernetes_cli"
""" client plugin_id for the metta kubernetes cli plugin """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class KubernetesCliPlugin(CliBase):
    """Fire command/group generator for various kubernetes plugin commands."""

    def fire(self) -> Dict[str, Any]:
        """Return command groups for Kubernetes plugins."""
        if (
            self._environment.fixtures().get(
                plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT,
                exception_if_missing=False,
            )
            is not None
        ):
            return {"kubernetes": KubernetesClientGroup(self._environment)}

        return {}


class KubernetesClientGroup:
    """Kubernetes client CLI commands."""

    def __init__(self, environment: Environment):
        """Add additional command groups for plugins and inject environment."""
        self._environment: Environment = environment

        if (
            self._environment.fixtures().get(
                plugin_id=METTA_PLUGIN_ID_KUBERNETES_YAML_WORKLOAD,
                exception_if_missing=False,
            )
            is not None
        ):
            self.yaml = KubernetesYamlWorkloadGroup(self._environment)
        if (
            self._environment.fixtures().get(
                plugin_id=METTA_PLUGIN_ID_KUBERNETES_HELM_WORKLOAD,
                exception_if_missing=False,
            )
            is not None
        ):
            self.helm = KubernetesHelmWorkloadGroup(self._environment)
        if (
            self._environment.fixtures().get(
                plugin_id=METTA_PLUGIN_ID_KUBERNETES_DEPLOYMENT_WORKLOAD,
                exception_if_missing=False,
            )
            is not None
        ):
            self.deployment = KubernetesDeploymentWorkloadGroup(self._environment)

    def _select_client(self, instance_id: str = "") -> Fixtures:
        """Pick a matching client."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT,
                instance_id=instance_id,
            )

        # Get the highest priority workload
        return self._environment.fixtures().get(plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT)

    def info(self, workload: str = "", deep: bool = False):
        """Get info about a client plugin."""
        fixture = self._select_client(instance_id=workload)
        return cli_output(fixture.info(deep=deep))

    def health(self, client: str = ""):
        """Retrieve the health status from the instance."""
        client_plugin = self._select_client(instance_id=client).plugin
        health = client_plugin.health()
        return cli_output(
            {
                "status": health.status(),
                "messages": list(health.messages()),
            }
        )

    def nodes(self, client: str = "", node: int = None):
        """Get info about a client plugin."""
        client_plugin = self._select_client(instance_id=client).plugin
        nodes = client_plugin.nodes()
        if node is None:
            return cli_output(list(node.to_dict() for node in nodes))
        return cli_output(nodes[node].to_dict())

    def nodes_status(self, client: str = "", node: int = None):
        """Get info about a client plugin."""
        client_plugin = self._select_client(instance_id=client).plugin
        nodes = client_plugin.nodes()
        if node is None:
            return cli_output(list(node.status.to_dict() for node in nodes))
        return cli_output(nodes[node].status.to_dict())

    def nodes_sysinfo(self, client: str = "", node: int = None):
        """Get info about a client plugin."""
        client_plugin = self._select_client(instance_id=client).plugin
        nodes = client_plugin.nodes()
        if node is None:
            return cli_output(list(node.status.node_info.to_dict() for node in nodes))
        return cli_output(nodes[node].status.node_info.to_dict())

    def readyz(self, client: str = ""):
        """Get kubernetes readiness info from the plugin."""
        client_plugin = self._select_client(instance_id=client).plugin
        try:
            return cli_output(client_plugin.readyz())

        except Exception as err:
            raise RuntimeError("Kubernetes is not ready") from err

    def livez(self, client: str = ""):
        """Get kubernetes livez info from the plugin."""
        client_plugin = self._select_client(instance_id=client).plugin
        try:
            return cli_output(client_plugin.livez())

        except Exception as err:
            raise RuntimeError("Kubernetes is not ready") from err

    def connect_service_proxy(self, namespace: str, service: str, client: str = ""):
        """Create a service proxy."""
        client_plugin = self._select_client(instance_id=client).plugin
        try:
            # Thank you kubernetes for the naming pattern
            # pylint: disable=invalid-name
            CoreV1Api = client_plugin.get_api("CoreV1Api")
            sc = CoreV1Api.connect_post_namespaced_service_proxy(namespace=namespace, name=service)

            return cli_output(sc)

        except Exception as err:
            raise RuntimeError("Exception trying to open the service proxy") from err


class KubernetesYamlWorkloadGroup:
    """Base Fire command group for terraform yaml workload plugin cli commands."""

    def __init__(self, environment: Environment):
        """Inject environment into command group object."""
        self._environment: Environment = environment

    def _select_fixture(self, instance_id: str = "") -> Fixtures:
        """Pick a matching workload fixture."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=METTA_PLUGIN_ID_KUBERNETES_YAML_WORKLOAD,
                instance_id=instance_id,
            )

        # Get the highest priority workload
        return self._environment.fixtures().get(plugin_id=METTA_PLUGIN_ID_KUBERNETES_YAML_WORKLOAD)

    def info(self, workload: str = "", deep: bool = False):
        """Get info about a yaml workload plugin."""
        fixture = self._select_fixture(instance_id=workload)
        fixture.plugin.prepare()

        return cli_output(fixture.info(deep=deep))

    def apply(self, workload: str = ""):
        """Run workload apply."""
        workload_plugin = self._select_fixture(instance_id=workload).plugin
        workload_plugin.prepare()

        objects = workload_plugin.apply()

        return cli_output(objects)

    def destroy(self, workload: str = ""):
        """Run workload destroy."""
        workload_plugin = self._select_fixture(instance_id=workload).plugin
        workload_plugin.prepare()

        destroy = workload_plugin.destroy()

        return cli_output(destroy)


class KubernetesHelmWorkloadGroup:
    """Base Fire command group for terraform helm workload cli commands."""

    def __init__(self, environment: Environment):
        """Inject environment into command group."""
        self._environment: Environment = environment

    def _select_fixture(self, instance_id: str = ""):
        """Pick a matching workload fixture."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=METTA_PLUGIN_ID_KUBERNETES_HELM_WORKLOAD,
                instance_id=instance_id,
            )

        # Get the highest priority workload
        return self._environment.fixtures().get(plugin_id=METTA_PLUGIN_ID_KUBERNETES_HELM_WORKLOAD)

    def info(self, workload: str = "", deep: bool = False) -> str:
        """Get info about a helm workload plugin."""
        fixture = self._select_fixture(instance_id=workload)
        fixture.plugin.prepare()

        return cli_output(fixture.info(deep=deep))

    def health(self, workload: str = "") -> str:
        """Get health status info about a helm workload plugin."""
        workload_plugin = self._select_fixture(instance_id=workload).plugin
        workload_plugin.prepare()

        return cli_output(workload_plugin.health())

    def apply(self, workload: str = "", wait: bool = True, debug: bool = False) -> str:
        """Run helm workload apply."""
        workload_plugin = self._select_fixture(instance_id=workload).plugin
        workload_plugin.prepare()

        objects = workload_plugin.apply(wait=wait, debug=debug)

        return cli_output(objects)

    def destroy(self, workload: str = "", debug: bool = False) -> str:
        """Run helm workload destroy."""
        workload_plugin = self._select_fixture(instance_id=workload).plugin
        workload_plugin.prepare()

        destroy = workload_plugin.destroy(debug=debug)

        return cli_output(destroy)

    def status(self, workload: str = "") -> str:
        """Run helm workload destroy."""
        workload_plugin = self._select_fixture(instance_id=workload).plugin
        workload_plugin.prepare()

        try:
            status = workload_plugin.status()
            status_info = {
                "name": status.name,
                "version": status.version,
                "namespace": status.namespace,
                "deleted": status.deleted,
                "description": status.description,
                "status": status.status,
                "config": status.config,
                "manifest": status.manifest,
            }
            return cli_output(status_info)
        except CalledProcessError:
            return cli_output("No status retrievable.")


class KubernetesDeploymentWorkloadGroup:
    """Base Fire command group for terraform deployment workload plugin cli commands."""

    def __init__(self, environment: Environment):
        """Inject environment into command group object."""
        self._environment: Environment = environment

    def _select_fixture(self, instance_id: str = "") -> Fixtures:
        """Pick a matching workload fixture."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=METTA_PLUGIN_ID_KUBERNETES_DEPLOYMENT_WORKLOAD,
                instance_id=instance_id,
            )

        # Get the highest priority workload
        return self._environment.fixtures().get(
            plugin_id=METTA_PLUGIN_ID_KUBERNETES_DEPLOYMENT_WORKLOAD
        )

    def info(self, workload: str = "", deep: bool = False):
        """Get info about workload plugin."""
        fixture = self._select_fixture(instance_id=workload)
        fixture.plugin.prepare()

        return cli_output(fixture.info(deep=deep))

    def health(self, workload: str = ""):
        """Retrieve the results from the deployment workload instance."""
        workload_plugin = self._select_fixture(instance_id=workload).plugin
        workload_plugin.prepare()
        health = workload_plugin.health()

        return cli_output(
            {
                "status": health.status(),
                "messages": list(health.messages()),
            }
        )

    def apply(self, workload: str = ""):
        """Run workload apply."""
        workload_plugin = self._select_fixture(instance_id=workload).plugin
        workload_plugin.prepare()

        objects = workload_plugin.apply()

        return cli_output(objects)

    def destroy(self, workload: str = ""):
        """Run workload destroy."""
        workload_plugin = self._select_fixture(instance_id=workload).plugin
        workload_plugin.prepare()

        destroy = workload_plugin.destroy()

        if destroy is None:
            return "workload not found."
        return cli_output(destroy)



"""

Common METTA plugins and functionality.

Package for common shared Metta plugins that can be used by various
other plugins as a based.

"""
from logging import getLogger
from typing import Dict, Any

from configerus import Config
from configerus.loaded import LOADED_KEY_ROOT
from configerus.plugin import FormatFactory

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.provisioner import METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER
from mirantis.testing.metta.workload import METTA_PLUGIN_INTERFACE_ROLE_WORKLOAD
from mirantis.testing.metta.output import METTA_PLUGIN_INTERFACE_ROLE_OUTPUT
from mirantis.testing.metta_cli.base import METTA_PLUGIN_INTERFACE_ROLE_CLI

from .common_config import add_common_config
from .dict_output import DictOutputPlugin, METTA_PLUGIN_ID_OUTPUT_DICT
from .text_output import TextOutputPlugin, METTA_PLUGIN_ID_OUTPUT_TEXT
from .combo_provisioner import (
    ComboProvisionerPlugin,
    METTA_PLUGIN_ID_PROVISIONER_COMBO,
    COMBO_PROVISIONER_CONFIG_LABEL,
)
from .config_format_output import ConfigFormatOutputPlugin, PLUGIN_ID_FORMAT_OUTPUT
from .user_cli import UserCliPlugin, METTA_PLUGIN_ID_CLI_USER


logger = getLogger("metta_common:init")


@Factory(
    plugin_id=METTA_PLUGIN_ID_OUTPUT_DICT,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_OUTPUT],
)
def metta_plugin_factory_output_dict(
    environment: Environment,
    instance_id: str,
    data: Dict = None,
    validator: str = "",
):
    """Create an output dict plugin."""
    return DictOutputPlugin(environment, instance_id, data, validator)


@Factory(
    plugin_id=METTA_PLUGIN_ID_OUTPUT_TEXT,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_OUTPUT],
)
def metta_plugin_factory_output_text(
    environment: Environment, instance_id: str = "", text: str = ""
):
    """Create an output text plugin."""
    return TextOutputPlugin(environment, instance_id, text)


@Factory(
    plugin_id=METTA_PLUGIN_ID_PROVISIONER_COMBO,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER],
)
def metta_plugin_factory_provisioner_combo(
    environment: Environment,
    instance_id: str,
    label: str = COMBO_PROVISIONER_CONFIG_LABEL,
    base: Any = LOADED_KEY_ROOT,
):
    """Create a provisioner combo plugin."""
    return ComboProvisionerPlugin(environment, instance_id, label=label, base=base)


# ----- Congiferus formatter for metta output plugins -----


@FormatFactory(plugin_id=PLUGIN_ID_FORMAT_OUTPUT)
def plugin_factory_format_output(config: Config, instance_id: str = ""):
    """Create an format plugin which replaces from output contents."""
    return ConfigFormatOutputPlugin(config, instance_id)


# ----- metta user cli plugin -----


@Factory(plugin_id=METTA_PLUGIN_ID_CLI_USER, interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLI])
def metta_plugin_factory_user_config(environment: Environment, instance_id: str = ""):
    """Create a user cli plugin."""
    return UserCliPlugin(environment, instance_id)


# ----- METTA bootstraps that we will use on config objects -----


# pylint: disable=unused-argument
def bootstrap_bootstrapper(config: Config):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    config (Config) : A config object that can be modified.
    """


# pylint: disable=unused-argument
def bootstrap_environment(environment: Environment):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment which should have validation config added to.
    """


def bootstrap_environment_common(environment: Environment):
    """Metta configerus bootstrap.

    Add some common Mirantis specific config options

    Add some common configerus sources for common data and common config source
    paths. Some of the added config is dynamic interpretation of environment,
    while also some default config paths are added if they can be interpeted

    @see .config.add_common_config() for details

    Also add the output formatter plugin to the environment config
    """
    add_common_config(environment)

    # Add a configerus output formatter which can interpret environemnt
    # outputs.
    instance_id = PLUGIN_ID_FORMAT_OUTPUT
    config: Config = environment.config()
    if config.has_formatter(instance_id):
        logger.debug("Existing output formatter for %s", environment.instance_id())
        output_formatter = environment.config().plugins[PLUGIN_ID_FORMAT_OUTPUT]
    else:
        logger.debug("New output formatter for %s", environment.instance_id())
        output_formatter = config.add_formatter(
            plugin_id=PLUGIN_ID_FORMAT_OUTPUT,
            instance_id=instance_id,
            priority=40,
        )
    output_formatter.set_environment(environment)



"""

Metta client plugin for kubernetes.

This client makes a strong effort to behave as an extension of the community
kubernetes python library, combining this with the metta config system, and
allowing creation by any provisioning source.


"""
import logging
import re
import time
from typing import Dict, Any, List

import kubernetes
from kubernetes.client import models
from kubernetes.client import api

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta_health.healthcheck import Health, HealthStatus

logger = logging.getLogger("metta.contrib.kubernetes.client.kubeapi")

METTA_PLUGIN_ID_KUBERNETES_CLIENT = "metta_kubernetes_kubeapi_client"
""" client plugin_id for the metta kubernetes client plugin """


KUBEAPI_CLIENT_INTERPRET_Z_REGEX = re.compile(
    r"^\[(?P<symbol>[+-])\](?P<name>\S+)\s{1}(?P<ok>\w+)$"
)


class KubernetesApiClientPlugin:
    """Metta Client plugin for Kubernetes.

    Construction:
    -------------
    As an metta Plugin, the class constructor is used to initialize the
    instance as a plugin.  This means that we lose the constructor, but we gain
    access to an metta.config.Config object.

    We use .args() to configure the instance and create the client, as per the
    client plugin standard.

    To use:
    -------

    2. Ask metta.client.make_client for an instance of the plugin
        a. it is more expected that a provisioner will provide the client
           directly
        b. you can ask for the plugin directly from metta.plugin.get_plugin()
        c. you can use metta.new_clients_from_config() if you have some
           configuration with the needed args, and it will build clients for
           you
    3. Configure using .args() adding a kubeconfig file with activated context
       if that wasn't done for you when you received the client
    4. Ask the client for a kubernetes API version client
       (such as get_api('CoreV1Api') => CoreV1Api)
    5. Use the kubernetes API client as normal

    ```
    import mirantis.testing.metta as metta
    import mirantis.testing.metta.client as metta_client
    import mirantis.testing.metta_kubernetes as metta_kubernetes

    config = metta.new_config()
    client = metta_client.make_client(
                metta_kubernetes.METTA_PLUGIN_ID_KUBERNETES_CLIENT,
                config, 'my-k8-instance')

    client.args('path/to/k8file')

    core_v1 = client.get_api('CoreV1Api')
    ns = core_v1.read_namespace(name="kube-system")
    print(f"NS: {ns}")
    ```

    Why use this:
    -------------
    It is cumbersome to use this if you already have access to the kubeconfig,
    but it fits into the plugin system, and therefore auto-loading and
    auto-configuring is possible with this implementation.
    Effectively, it is not easier to use this plu-gin over the native K8 client
    unless you consider that a provisioner can provide you this plugin already
    configured with its own information.

    This plugin is particularly usefull as it can be used for the helm and
    deployment workload plugins directly, which know how to use it to apply
    workloads to a kubernetes cluster.

    """

    def __init__(self, environment: Environment, instance_id: str, kube_config_file: str = ""):
        """Run the super constructor but also set class properties.

        This implements the args part of the client interface.

        Here we expect to receive a path to a KUBECONFIG file with a context set
        and we create a Kubernetes client for replace_existing=Trueuse.  After
        that this can provide Core api clients as per the kubernetes SDK

        Parameters:
        -----------
        config_file (str): String path to the kubernetes config file to use

        """
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        logger.debug("Creating Kuberentes client from config file")
        self._api_client = kubernetes.config.new_client_from_config(config_file=kube_config_file)
        """ Kubernetes api client """

        self.config_file = kube_config_file
        """ Kube config file, in case you need to steal it. """

    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Return dict data about this plugin for introspection."""
        info: Dict[str, Any] = {"kubernetes": {"config_file": self.config_file}}

        if deep:
            try:
                info["nodes"] = list(node.to_dict() for node in self.nodes())
            # pylint: disable=broad-except
            except Exception:
                # Still continue to run
                pass

        return info

    def get_api(self, name: str):
        """Get an kubernetes API."""
        if hasattr(kubernetes.client, name):
            return getattr(kubernetes.client, name)(self._api_client)

        raise KeyError(f"Unknown API requested: {api}")

    def utils_create_from_yaml(self, yaml_file: str, **kwargs):
        """Run a kube apply from a yaml file."""
        return kubernetes.utils.create_from_yaml(
            k8s_client=self._api_client, yaml_file=yaml_file, **kwargs
        )

    def utils_create_from_dict(self, data: Dict, **kwargs):
        """Run a kube apply from dict of K8S yaml."""
        return kubernetes.utils.create_from_dict(k8s_client=self._api_client, data=data, **kwargs)

    def nodes(self) -> List[models.v1_node.V1Node]:
        """Return V1Node list.

        Returns:
        --------
        List of kubernetes cluster nodes as V1Node objects.

        You can get node status from node.status.

        """
        node_items = self.get_api("CoreV1Api").list_node().items

        return node_items

    def kubelet_ready_wait(self, timeout: int = 30, period: int = 1):
        """Wait until all nodes' kubelets are ready.

        Some operations need all k8s nodes to be ready or they will fail.
        In such cases the readyz approach does not suffice, as it only
        guarantees that enough nodes are online to allow api calls to
        work.

        """
        err = None
        while timeout > 0:
            try:
                for node in self.nodes():
                    kubelet_condition = node_status_condition(node, "KubeletReady")

                    if not kubelet_condition.status == "True":
                        raise RuntimeError(f"Node kubelet is not ready: {node.metadata.name}")
                return True

            except RuntimeError as this_err:
                logger.debug("node kubelet not ready: %s", this_err)
                err = this_err
                time.sleep(period)
                timeout = timeout - period
                continue

        raise RuntimeError("Timed out waiting for kubernetes to become ready") from err

    def readyz_wait(self, timeout: int = 30, period: int = 1):
        """Wait until kubernetes is ready before returning."""
        err = None
        while timeout > 0:
            try:
                ready = self.readyz()
                return ready
            except kubernetes.client.rest.ApiException as this_err:
                err = this_err
                time.sleep(period)
                timeout = timeout - period
                continue

        raise RuntimeError("Timed out waiting for kubernetes to become ready") from err

    def readyz(self):
        """Check the general readyz endpoint.

        Returns:
        --------
        Dict of service:status_dict with values symbol (+) and string ok value

        Raises:
        -------
        Will raise an exception is kubernetes isn't avaialble

        """
        return self._interpret_z_response("/readyz", params={"verbose": "true"})

    def livez(self):
        """Check the general livez endpoint.

        Returns:
        --------
        Dict of service:status_dict with values symbol (+) and string ok value

        Raises:
        -------
        Will raise an exception is kubernetes isn't available

        """
        return self._interpret_z_response("/livez", params={"verbose": "true"})

    def _interpret_z_response(
        self, endpoint: str, method: str = "GET", params: Dict[str, str] = None
    ) -> Dict[str, Dict[str, str]]:
        """Interpret that readyz/livez response format into a dict.

        because both livez and readyz return similar a machine un-friendly response, we use this
        method to retrieve and interpret the result.

        """
        # this will produce an exception if K8s is not ready
        response = self._api_client.call_api(
            method=method,
            resource_path=endpoint,
            query_params=params if params is not None else {},
            _preload_content=False,
        )[0]

        interpreted = {}
        for line in response.read(cache_content=False).decode("utf-8").split("\n"):
            match = KUBEAPI_CLIENT_INTERPRET_Z_REGEX.fullmatch(line)
            if match:
                interpreted[match.group("name")] = {
                    "symbol": match.group("symbol"),
                    "ok": match.group("ok"),
                }

        return interpreted

    # this allows consumers to get a watch without importing kubernetes
    # pylint: disable=no-self-use
    def watch(self):
        """Get a kubernetes watch instance."""
        return kubernetes.watch.Watch()

    # healthcheck interfaces

    def health(self) -> Health:
        """Determine the health of the K8s instance."""
        k8s_health = Health(source=self._instance_id, status=HealthStatus.UNKNOWN)

        for test_health_function in [
            self._health_k8s_readyz,
            self._health_k8s_livez,
            self._health_k8s_node_health,
            self._health_k8s_alldeployment_health,
            self._health_k8s_alldaemonset_health,
            self._health_k8s_allstatefulset_health,
            self._health_k8s_allpod_health,
        ]:
            try:
                test_health = test_health_function()
            # pylint: disable=broad-except
            except Exception as err:
                test_health = Health(source=self._instance_id)
                test_health.critical(f"{test_health_function} exception: {err}")
            finally:
                k8s_health.merge(test_health)
        return k8s_health

    def _health_k8s_readyz(self) -> Health:
        """Check if kubernetes thinks the pod is healthy."""
        health = Health(source=self._instance_id)

        try:
            if self.readyz():
                health.healthy("KubeAPI: readyz reports ready")
            else:
                health.warning("KubeAPI: readyz reports NOT ready.")
        # pylint: disable=broad-except
        except Exception as err:
            health.error(f"Could not retrieve readyz: {err}")

        return health

    def _health_k8s_livez(self) -> Health:
        """Check if kubernetes thinks the pod is healthy."""
        health = Health(source=self._instance_id)

        try:
            if self.livez():
                health.healthy("KubeAPI: livez reports live")
            else:
                health.warning("KubeAPI: livez reports NOT live.")
        # pylint: disable=broad-except
        except Exception as err:
            health.error(f"Could not retrieve livez: {err}")

        return health

    def _health_k8s_node_health(self) -> Health:
        """Check if kubernetes thinks the nodes are healthy."""
        health = Health(source=self._instance_id)

        try:
            for node in self.nodes():
                name = node.metadata.name
                no_issues = True

                condition = next(
                    (
                        condition
                        for condition in node.status.conditions
                        if condition.type == "Ready"
                    ),
                    None,
                )
                if condition is not None and condition.status != "True":
                    health.warning(f"KubeAPI: {name}: {condition.message}")
                    no_issues = False

                condition = next(
                    (
                        condition
                        for condition in node.status.conditions
                        if condition.type == "NetworkUnavailable"
                    ),
                    None,
                )
                if condition is not None and condition.status == "True":
                    health.warning(f"KubeAPI: {name}: {condition.message}")
                    no_issues = False

                condition = next(
                    (
                        condition
                        for condition in node.status.conditions
                        if condition.type == "MemoryPressure"
                    ),
                    None,
                )
                if condition is not None and condition.status == "True":
                    health.warning(f"KubeAPI: {name}: {condition.message}")
                    no_issues = False

                condition = next(
                    (
                        condition
                        for condition in node.status.conditions
                        if condition.type == "DiskPressure"
                    ),
                    None,
                )
                if condition is not None and condition.status == "True":
                    health.warning(f"KubeAPI: {name}: {condition.message}")
                    no_issues = False

                condition = next(
                    (
                        condition
                        for condition in node.status.conditions
                        if condition.type == "PIDPressure"
                    ),
                    None,
                )
                if condition is not None and condition.status == "True":
                    health.warning(f"KubeAPI: {name}: {condition.message}")
                    no_issues = False

                if no_issues:
                    health.healthy(f"KubeAPI: Node {name} reports healthy.")
                else:
                    health.error(f"KubeAPI: Node {name} reporting issues.")

        # pylint: disable=broad-except
        except Exception as err:
            health.error(f"KubeAPI:Exception occured when check kubelet health: {err}")

        return health

    # pylint: disable=too-many-branches
    def _health_k8s_alldeployment_health(self) -> Health:
        """Check if kubernetes thinks all the deployments are healthy."""
        health = Health(source=self._instance_id)

        apps_v1_api: api.apps_v1_api.AppsV1Api = self.get_api("AppsV1Api")

        unhealthy_dep_count = 0
        # pylint: disable=no-member
        for deployment in apps_v1_api.list_deployment_for_all_namespaces().items:
            namespace = deployment.metadata.namespace
            name = deployment.metadata.name

            no_issues = True

            if not deployment.status.conditions:
                health.unknown(
                    f"KubeAPI:Deployment: [{namespace}/{name}] "
                    "Deployment does not have any conditions (yet?)"
                )
                continue

            available_condition = next(
                (
                    condition
                    for condition in deployment.status.conditions
                    if condition.type == "Available"
                ),
                None,
            )
            progressing_condition = next(
                (
                    condition
                    for condition in deployment.status.conditions
                    if condition.type == "Progressing"
                ),
                None,
            )
            if available_condition and available_condition.status == "True":
                pass
            elif progressing_condition and progressing_condition.status == "True":
                health.warning(
                    f"KubeAPI:Deployment: [{namespace}/{name}] "
                    "Deployment is progressing "
                    f"-> {progressing_condition.message}"
                )
                no_issues = False
            else:
                messages = "\n".join(
                    list(
                        condition.message
                        for condition in [progressing_condition, available_condition]
                        if condition is not None
                    )
                )
                health.warning(
                    f"KubeAPI:Deployment: [{namespace}/{name}] "
                    "Deployment is neither progressing nor available "
                    f"-> {messages}"
                )
                no_issues = False

            for condition in deployment.status.conditions:
                if condition.type in ["Available", "Progressing"]:
                    pass

                elif condition.status != "True":
                    health.warning(
                        f"KubeAPI:Deployment: [{namespace}/{name}] {condition.type} "
                        f"-> {condition.message}"
                    )
                    no_issues = False

            if no_issues:
                health.healthy(f"KubeAPI:Deployment: [{namespace}/{name}] is healthy")
            else:
                health.warning(f"KubeAPI:Deployment: [{namespace}/{name}] is not healthy")
                unhealthy_dep_count += 1

        if unhealthy_dep_count == 0:
            health.healthy("KubeAPI: all deployments report healthy")
        elif unhealthy_dep_count < 3:
            health.warning("KubeAPI: some deployments report condition failures")
        else:
            health.error("KubeAPI: Kubernetes Reports cluster is unhealthy (deployment health)")

        return health

    def _health_k8s_alldaemonset_health(self) -> Health:
        """Check if kubernetes thinks all the daemonsets are healthy."""
        health = Health(source=self._instance_id)

        apps_v1_api: api.apps_v1_api.AppsV1Api = self.get_api("AppsV1Api")

        unhealthy_dae_count = 0
        # pylint: disable=no-member
        for daemonset in apps_v1_api.list_daemon_set_for_all_namespaces().items:
            namespace = daemonset.metadata.namespace
            name = daemonset.metadata.name
            status = daemonset.status

            if status.collision_count is not None and status.collision_count > 0:
                health.warning(
                    f"Daemonset: [{namespace}/{name}] collision_count "
                    "-> Reports some collisions: "
                    f"{status.collision_count}"
                )
                unhealthy_dae_count += 1
            if status.number_unavailable is not None and status.number_unavailable > 0:
                health.warning(
                    f"Daemonset: [{namespace}/{name}] number_unavailable "
                    "-> Reports some unavailable pods: "
                    f"{status.number_unavailable}"
                )
                unhealthy_dae_count += 1
            if status.desired_number_scheduled < status.current_number_scheduled:
                health.warning(
                    f"Daemonset: [{namespace}/{name}] desired_number_scheduled "
                    "-> Does not have the desired number scheduled: "
                    f"{status.desired_number_scheduled} < "
                    f"{status.current_number_scheduled}"
                )
                unhealthy_dae_count += 1

            if status.conditions:
                for condition in status.conditions:
                    if condition.status == "True":
                        health.healthy(
                            f"Daemonset: [{namespace}/{name}] {condition.type} "
                            f"-> {condition.message}"
                        )
                    else:
                        health.warning(
                            f"Daemonset: [{namespace}/{name}] {condition.type} "
                            f"-> {condition.message}"
                        )
                        unhealthy_dae_count += 1

        if unhealthy_dae_count == 0:
            health.healthy("KubeAPI: all daemonsets report as healthy")
        elif unhealthy_dae_count < 3:
            health.warning("KubeAPI: some daemonsets report condition failures")
        else:
            health.error("KubeAPI: Kubernetes Reports cluster is unhealthy (daemonset health)")

        return health

    def _health_k8s_allstatefulset_health(self) -> Health:
        """Check if kubernetes thinks all the statefulsets are healthy."""
        health = Health(source=self._instance_id)

        apps_v1_api: api.apps_v1_api.AppsV1Api = self.get_api("AppsV1Api")

        unhealthy_count = 0
        # pylint: disable=no-member
        for statefulset in apps_v1_api.list_stateful_set_for_all_namespaces().items:
            namespace = statefulset.metadata.namespace
            name = statefulset.metadata.name
            status = statefulset.status

            # {'collision_count': 0,
            #  'conditions': None,
            #  'current_replicas': 1,
            #  'current_revision': 'loki-workload-67877b465c',
            #  'observed_generation': 1,
            #  'ready_replicas': 1,
            #  'replicas': 1,
            #  'update_revision': 'loki-workload-67877b465c',
            #  'updated_replicas': 1}

            if status.collision_count is not None and status.collision_count > 0:
                health.warning(
                    f"KubeAPI:Statefulset: [{namespace}/{name}] "
                    "-> Reports some collisions: "
                    f"{status.collision_count}"
                )
                unhealthy_count += 1

            if status.conditions:
                for condition in status.conditions:
                    if condition.status == "True":
                        health.healthy(
                            f"KubeAPI:Statefulset: [{namespace}/{name}] {condition.type} "
                            f"-> {condition.message}"
                        )
                    else:
                        health.warning(
                            f"KubeAPI:Statefulset: [{namespace}/{name}] {condition.type} "
                            f"-> {condition.message}"
                        )
                        unhealthy_count += 1

        if unhealthy_count == 0:
            health.healthy("KubeAPI: all statefulsets report as healthy")
        elif unhealthy_count < 3:
            health.warning("KubeAPI: some statefulsets report condition failures")
        else:
            health.error("KubeAPI: Kubernetes Reports cluster is unhealthy (statefulset health)")

        return health

    def _health_k8s_allpod_health(self) -> Health:
        """Check if kubernetes thinks all the pods are healthy."""
        health = Health(source=self._instance_id)

        core_v1_api = self.get_api("CoreV1Api")

        unhealthy_pod_count = 0
        for pod in core_v1_api.list_pod_for_all_namespaces().items:
            if pod.status.phase == "Failed":
                health.error(f"KubeAPI: pod failed: {pod.metadata.name}")
                unhealthy_pod_count += 1
        if unhealthy_pod_count == 0:
            health.healthy("KubeAPI: all pods report as healthy")
        elif unhealthy_pod_count < 2:
            health.warning("KubeAPI: some pods report as failed")
        else:
            health.error("KubeAPI: Kubernetes Reports cluster is unhealthy (pod health)")

        return health


def node_status_condition(
    node: models.v1_node.V1Node, cond_reason: str, cond_type: str = "Ready"
) -> models.v1_node_condition.V1NodeCondition:
    """Retrieve a status condition of matching reason from a node."""
    status: models.v1_node_status.V1NodeStatus = node.status

    for condition in status.conditions:
        if condition.reason == cond_reason and condition.type == cond_type:
            return condition
    raise RuntimeError(f"No matching condition found for {cond_reason}/{cond_type}")



"""

Metta testkit functionality.

this module registers the testkit metta plugin factories to make them available
for use.

"""
from typing import Any, Dict

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.provisioner import METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER
from mirantis.testing.metta.client import METTA_PLUGIN_INTERFACE_ROLE_CLIENT
from mirantis.testing.metta_cli.base import METTA_PLUGIN_INTERFACE_ROLE_CLI

from .provisioner import (
    TestkitProvisionerPlugin,
    METTA_TESTKIT_PROVISIONER_PLUGIN_ID,
    TESTKIT_PROVISIONER_CONFIG_LABEL,
    TESTKIT_PROVISIONER_CONFIG_BASE,
)
from .client import (
    TestkitClientPlugin,
    METTA_TESTKIT_CLIENT_PLUGIN_ID,
)
from .cli import TestkitCliPlugin, METTA_TESTKIT_CLI_PLUGIN_ID


@Factory(
    plugin_id=METTA_TESTKIT_PROVISIONER_PLUGIN_ID,
    interfaces=[
        METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER,
        METTA_TESTKIT_PROVISIONER_PLUGIN_ID,
    ],
)
def metta_plugin_factory_provisioner_testkit(
    environment: Environment,
    instance_id: str,
    label: str = TESTKIT_PROVISIONER_CONFIG_LABEL,
    base: Any = TESTKIT_PROVISIONER_CONFIG_BASE,
) -> TestkitProvisionerPlugin:
    """Create an metta testkit provisioner plugin."""
    return TestkitProvisionerPlugin(environment, instance_id, label=label, base=base)


@Factory(
    plugin_id=METTA_TESTKIT_CLIENT_PLUGIN_ID,
    interfaces=[
        METTA_PLUGIN_INTERFACE_ROLE_CLIENT,
        METTA_TESTKIT_CLIENT_PLUGIN_ID,
    ],
)
def metta_plugin_factory_client_testkit(
    environment: Environment,
    instance_id: str,
    system_name: str,
    config_file: str,
    systems: Dict[str, Dict[str, str]] = None,
) -> TestkitClientPlugin:
    """Create an metta testkit client plugin."""
    return TestkitClientPlugin(
        environment, instance_id, system_name=system_name, config_file=config_file, systems=systems
    )


@Factory(
    plugin_id=METTA_TESTKIT_CLI_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLI],
)
def metta_plugin_factory_cli_testkit(
    environment: Environment, instance_id: str = ""
) -> TestkitCliPlugin:
    """Create an litmuschaos cli plugin."""
    return TestkitCliPlugin(environment, instance_id)


# ----- SetupTools EntryPoint METTA BootStrapping -----


# pylint: disable=unused-argument
def bootstrap_environment(environment):
    """METTA_testkit bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment

    """



"""

Interact and manipulate Configerus objects.

"""
import logging
from typing import Any, List, Union

from configerus.config import Config
from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.files import PLUGIN_ID_SOURCE_PATH, CONFIGERUS_PATH_KEY
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT, CONFIGERUS_DICT_DATA_KEY
from configerus.contrib.env import (
    PLUGIN_ID_SOURCE_ENV_SPECIFIC,
    CONFIGERUS_ENV_SPECIFIC_BASE_KEY,
    PLUGIN_ID_SOURCE_ENV_JSON,
    CONFIGERUS_ENV_JSON_ENV_KEY,
)

from .globals import global_config

logger = logging.getLogger("metta.config")

DEFAULT_SOURCE_PRIORITY = 75
"""Default source plugin priority if not explicitly given."""

METTA_CONFIG_CONFIG_SOURCE_KEY = "config.sources"
"""Configerus config .get() key used to find config source definitions."""


def default_config(additional_bootstraps: List[str] = None):
    """Provide a global config object, creating one if needed."""
    config = global_config

    if additional_bootstraps is not None:
        for bootstrap_entrypoint_id in additional_bootstraps:
            logger.debug("bootstrapping: %s", bootstrap_entrypoint_id)
            config.bootstrap(bootstrap_entrypoint_id)

    return config


def add_config_sources_from_config(
    config: Config,
    label: str = "config",
    base: Union[str, List[Any]] = LOADED_KEY_ROOT,
    default_source_priority: int = DEFAULT_SOURCE_PRIORITY,
):
    """Add more config sources based on in config settings.
    Read some config which will tell us where more config can be found.
    This lets us use config to extend config, and is what make metta
    entirely extensible from a single metta.yml file.
    In the trade-off battle between configurable and convention, this leans
    heavily towards configuration, but it easily lends to standards.
    Parameters:
    -----------
    label (str) : configurus load label.
    base (str|List[str]) : configerus get key as a base for retrieving all
        config settings.
    """
    config_environment = config.load(label)
    config_sources = config_environment.get(base, default={})

    for instance_id in config_sources.keys():
        instance_base = [base, instance_id]

        # Keep in mind that the following plugin metadata is about
        # configerus plugins, not metta plugins.

        plugin_id = config_environment.get([instance_base, "plugin_id"])
        priority = config_environment.get(
            [instance_base, "priority"], default=default_source_priority
        )

        logger.debug(
            "Adding metta sourced config plugin: %s:%s",
            plugin_id,
            instance_id,
        )
        plugin = config.add_source(plugin_id=plugin_id, instance_id=instance_id, priority=priority)

        # Configerus plugins all work differently so we take a different
        # approach per plugin
        if plugin_id == PLUGIN_ID_SOURCE_PATH:
            path = config_environment.get([instance_base, CONFIGERUS_PATH_KEY])
            plugin.set_path(path=path)
        elif plugin_id == PLUGIN_ID_SOURCE_DICT:
            data = config_environment.get([instance_base, CONFIGERUS_DICT_DATA_KEY])
            plugin.set_data(data=data)
        elif plugin_id == PLUGIN_ID_SOURCE_ENV_SPECIFIC:
            source_base = config_environment.get([instance_base, CONFIGERUS_ENV_SPECIFIC_BASE_KEY])
            plugin.set_base(base=source_base)
        elif plugin_id == PLUGIN_ID_SOURCE_ENV_JSON:
            source_env = config_environment.get([instance_base, CONFIGERUS_ENV_JSON_ENV_KEY])
            plugin.set_env(env=source_env)
        # this should probably be a configerus standard
        elif hasattr(plugin, "set_data"):
            data = config_environment.get([instance_base, "data"])
            plugin.set_data(data=data)
        else:
            logger.warning(
                "had no way of configuring new Configerus source plugin %s",
                plugin_id,
            )



"""

Kubernetes workload plugin.

This plugin uses the kube_api client to deploy workloads to kubernetes, from a deployment
definition from configuration.

"""

import logging
from typing import Any, Dict

import kubernetes

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures
from mirantis.testing.metta_health.healthcheck import Health, HealthStatus

from .kubeapi_client import KubernetesApiClientPlugin, METTA_PLUGIN_ID_KUBERNETES_CLIENT

logger = logging.getLogger("metta.contrib.kubernetes.workload.deployment")

METTA_PLUGIN_ID_KUBERNETES_DEPLOYMENT_WORKLOAD = "metta_kubernetes_deployment_workload"
""" workload plugin_id for the metta_kubernetes deployment plugin """

KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_LABEL = "kubernetes"
KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_BASE = "workload.deployment"

KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_NAMESPACE = "namespace"
KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_BODY = "body"


# pylint: disable=too-many-instance-attributes
class KubernetesDeploymentWorkloadPlugin:
    """Metta workload plugin for Kubernetes workload."""

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        label: str = KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_LABEL,
        base: Any = KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_BASE,
    ):
        """Run the super constructor but also set class properties.

        This implements the args part of the client interface.

        Here we expect to receive a path to a KUBECONFIG file with a context set
        and we create a Kubernetes client for use.  After that this can provide
        Core api clients as per the kubernetes SDK

        Parameters:
        -----------
        config_file (str): String path to the kubernetes config file to use

        """
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        self._config_label: str = label
        """ configerus load label that should contain all of the config """
        self._config_base: str = base
        """ configerus get key that should contain all tf config """

        workload_config = self._environment.config().load(self._config_label)

        self.name: str = workload_config.get(
            [
                self._config_base,
                KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_BODY,
                "metadata.name",
            ],
            default="not-declared",
        )
        """Kubernetes deployment name string."""
        self.namespace: str = workload_config.get(
            [self._config_base, KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_NAMESPACE]
        )
        """Kubernetes namespace to use for the deployment."""
        self.body: Dict[str, Any] = workload_config.get(
            [self._config_base, KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_BODY]
        )
        """Kubernetes deployment definition dict."""

        self._kubeapi_client: KubernetesApiClientPlugin = None
        """KubeAPI client for connecting to kubernetes."""

        self._deployment: object = None
        """KubeAPI get deployment result."""

        # do an initial prepare in case it is never properly run
        try:
            self.prepare()
        # pylint: disable=broad-except
        except Exception:
            pass
        # check if the deployment is already running
        self.read()

    # deep argument is an info() standard across plugins
    # pylint: disable=unused-argument
    def info(self, deep: bool = False):
        """Return dict data about this plugin for introspection."""
        workload_config = self._environment.config().load(self._config_label)

        if self._kubeapi_client is None:
            # let's take a stab at finding a client for declarative cases
            try:
                kubeclient = self._environment.fixtures().get_plugin(
                    plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT,
                )
                kubeclient_info = kubeclient.info()
            except KeyError:
                # we will just work around a missing kube api plugin
                kubeclient_info = None
        else:
            kubeclient_info = self._kubeapi_client.info()

        return {
            "workload": {
                "deployment": {
                    "namespace": workload_config.get(
                        [
                            self._config_base,
                            KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_NAMESPACE,
                        ]
                    ),
                    "body": workload_config.get(
                        [
                            self._config_base,
                            KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_BODY,
                        ]
                    ),
                },
                "required_fixtures": {
                    "kubernetes": {
                        "plugin_id": METTA_PLUGIN_ID_KUBERNETES_CLIENT,
                        "kube_client": kubeclient_info,
                    }
                },
            }
        }

    def prepare(self, fixtures: Fixtures = None):
        """Find the kubeapi client from a set of fixtures and retrieve config
        for running the workload.

        Parameters:
        ----------
        fixtures (Fixtures) : a set of fixtures that this workload will use to
            retrieve a kubernetes client plugin.

        """
        if self._deployment is not None:
            self.destroy()

        if fixtures is None:
            fixtures = self._environment.fixtures()

        try:
            self._kubeapi_client = fixtures.get_plugin(plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT)
        except KeyError as err:
            raise NotImplementedError(
                "Workload could not find the needed client: "
                f"{METTA_PLUGIN_ID_KUBERNETES_CLIENT}"
            ) from err

    def apply(self):
        """Run the workload."""
        apps_v1 = self._kubeapi_client.get_api("AppsV1Api")
        self._deployment = apps_v1.create_namespaced_deployment(
            body=self.body, namespace=self.namespace
        )
        return self._deployment

    def destroy(self):
        """Destroy any created resources."""
        body = kubernetes.client.V1DeleteOptions(
            propagation_policy="Foreground", grace_period_seconds=5
        )

        apps_v1 = self._kubeapi_client.get_api("AppsV1Api")
        status = apps_v1.delete_namespaced_deployment(
            name=self.name, namespace=self.namespace, body=body
        )

        self._deployment = None

        return status

    def read(self):
        """Retrieve the deployment job."""
        if self._kubeapi_client is None:
            return None

        apps_v1 = self._kubeapi_client.get_api("AppsV1Api")

        try:
            self._deployment = apps_v1.read_namespaced_deployment(
                name=self.name, namespace=self.namespace
            )
            return self._deployment
        except kubernetes.client.rest.ApiException:
            return None

    # INTERFACE: healthcheck
    def health(self) -> Health:
        """Determine the health of the K8s deployment."""
        dep_health = Health(source=self._instance_id, status=HealthStatus.UNKNOWN)

        if self._deployment is None:
            dep_health.info(f"Deployment: {self._instance_id} not yet started.")
            return dep_health

        for test_health_function in [self._health_deployment_status]:
            test_health = test_health_function()
            dep_health.merge(test_health)

        return dep_health

    def _health_deployment_status(self):
        """Check if kubernetes thinks the deployment is healthy."""
        health = Health(source=self._instance_id)

        apps_v1 = self._kubeapi_client.get_api("AppsV1Api")

        try:
            deployment = apps_v1.read_namespaced_deployment(self.name, self.namespace)
            status = deployment.status

            if status is None:
                health.error(f"Deployment: [{self.namespace}/{self.name}] retrieved no status.")
            if status.conditions is None:
                health.warning(
                    f"Deployment: [{self.namespace}/{self.name}] retrieved no status conditions."
                )
            else:

                available_condition = next(
                    condition for condition in status.conditions if condition.type == "Available"
                )
                progressing_condition = next(
                    condition for condition in status.conditions if condition.type == "Progressing"
                )
                if available_condition and available_condition.status == "True":
                    health.healthy(
                        f"Deployment: [{self.namespace}/{self.name}] "
                        "Deployment is available "
                        f"-> {available_condition.message}"
                    )
                elif progressing_condition and progressing_condition.status == "True":
                    health.warning(
                        f"Deployment: [{self.namespace}/{self.name}] "
                        "Deployment is progressing "
                        f"-> {progressing_condition.message}"
                    )
                else:
                    health.error(
                        f"Deployment: [{self.namespace}/{self.name}] "
                        "Deployment is neither progressing nor available "
                        f"-> {available_condition.message} && {progressing_condition.message}"
                    )

                for condition in status.conditions:
                    if condition.type in ["Available", "Progressing"]:
                        pass

                    elif condition.status == "True":
                        health.healthy(
                            f"Deployment: [{self.namespace}/{self.name}] {condition.type} "
                            f"-> {condition.message}"
                        )
                    else:
                        health.error(
                            f"Deployment: [{self.namespace}/{self.name}] {condition.type} "
                            f"-> {condition.message}"
                        )

        except kubernetes.client.rest.ApiException as err:
            health.error(f"Deployment: K8S REST API exception occured: {err}")

        return health



"""

Ansible Playbook client.

This module centralizes the ansible integration for metta, to be used by
various plugins for the package.  The client expects to run playbook style
queues, receiving task instructions in collection/primitive form.

"""

from typing import Dict, Any
import logging

import ansible.constants as C
from ansible.executor.task_queue_manager import TaskQueueManager
from ansible.inventory.manager import InventoryManager
from ansible.parsing.dataloader import DataLoader
from ansible.playbook.play import Play
from ansible.plugins.callback import CallbackBase
from ansible.vars.manager import VariableManager
from ansible.utils.context_objects import CLIArgs
from ansible import context

from .ansible_callback import ResultsCallback

logger = logging.getLogger("ansible")


class AnsiblePlay:
    """

    Ansible integration class.

    Allows execution of ansible commands using the ansible core directl.

    """

    def __init__(self, inventory_path: str, ansiblecfg_path: str = None):
        """Initial configuration of ansible plugin."""
        self.inventory_path: str = inventory_path
        """Path to the ansible inventory file."""

        # We should pass these in.
        self._passwords: Dict[str, str] = {}
        """Passwords that should be passed to the ansible queue exec."""

        # We should pass these in as args to the constructor
        # WE will likely want to add modules/libraries here.
        cliargs = {"connection": "smart", "verbosity": 10}

        # things we should add to the cliargs dict:
        # - module_path=[] : string paths that should provide ansible python modules
        # - check=False : I don't know what this does.
        # - diff=False : I think ansible outputs more cached diff stuff
        # Things we probably don't need:
        # - become_user=None / become_method=None : sudo kind of stuff.
        # - forks=10 : how many threads to run

        context.CLIARGS = CLIArgs(cliargs)

        if ansiblecfg_path:
            C.config._config_file = ansiblecfg_path
            C.config.update_config_data()
            logger.warning(
                "NOW OVER HERE: %s", C.config.get_config_value_and_origin("COLLECTIONS_PATHS")
            )

    def ping(self, hosts: Any = "all", gather_facts: bool = False) -> ResultsCallback:
        """Ping all hosts in the inventory."""
        # create data structure that represents our play, including tasks,
        # this is basically what our YAML loader does internally.
        play = dict(
            name="Hosts ping",
            hosts=hosts,
            gather_facts="yes" if gather_facts else "no",
            tasks=[
                dict(action=dict(module="ping")),
            ],
        )

        # Instantiate our custom callback for handling results as they come in.
        # Ansible expects this to be one of its main display outlets
        results_callback = ResultsCallback()

        self.play(play, callback=results_callback)
        return results_callback

    def debug(self, hosts: Any = "all", gather_facts: bool = False) -> ResultsCallback:
        """Debug all hosts in the inventory."""
        # create data structure that represents our play, including tasks,
        # this is basically what our YAML loader does internally.
        play = dict(
            name="Hosts ping",
            hosts=hosts,
            gather_facts="yes" if gather_facts else "no",
            tasks=[
                dict(action=dict(module="debug")),
            ],
        )

        # Instantiate our custom callback for handling results as they come in.
        # Ansible expects this to be one of its main display outlets
        results_callback = ResultsCallback()

        self.play(play, callback=results_callback)
        return results_callback

    def setup(self, hosts: Any = "all", gather_facts: bool = False) -> ResultsCallback:
        """Retrieve setup info from all hosts in the inventory."""
        play = dict(
            name="Hosts setup",
            hosts=hosts,
            gather_facts="yes" if gather_facts else "no",
            tasks=[
                dict(action=dict(module="setup")),
            ],
        )

        # Instantiate our custom callback for handling results as they come in.
        # Ansible expects this to be one of its main display outlets
        results_callback = ResultsCallback()

        self.play(play, callback=results_callback)
        return results_callback

    # ansible contants are not all defined before using.
    # pylint: disable=no-member
    def play(
        self,
        play: Dict[str, Any],
        callback: CallbackBase = C.DEFAULT_STDOUT_CALLBACK,
        variables: Dict[str, str] = None,
    ) -> int:
        """Run a playbook in a queue defined as a dict.

        Parameters:
        -----------
        play: dict representation of a yaml playbook file
        callback: callback option. If you want real data out of the callbacks then pass in an
            object which can capture the data.
            You can use this to send a custom object that you intend to use to retrieve results.

            @NOTE this looks like the wrong approach.  We should probably ensure that a custom
                callback is always included in the callback set and then always return it.

        Returns:
        --------
        Integer QueueManager result code

        """
        # initialize needed objects

        # Takes care of finding and reading yaml, json and ini files.
        loader = DataLoader()
        # Will handle our host/inventories.
        inventory = InventoryManager(loader=loader, sources=[self.inventory_path])
        # takes care of merging all the different sources to give you unified view of variable
        variable_manager = VariableManager(loader=loader, inventory=inventory)

        # instantiate task queue manager, which takes care of forking and setting up all objects to
        # iterate over host list and tasks
        # IMPORTANT: This also adds library dirs paths to the module loader
        # IMPORTANT: and so it must be initialized before calling `Play.load()`.
        tqm = TaskQueueManager(
            inventory=inventory,
            variable_manager=variable_manager,
            loader=loader,
            passwords=self._passwords,
            stdout_callback=callback,
        )

        # Create play object, playbook objects use .load instead of init or new methods,
        # this will also automatically create the task objects from info provided in play_source
        play_instance = Play().load(
            play,
            variable_manager=variable_manager,
            loader=loader,
            vars=variables if variables else {},
        )

        # Actually run it
        try:
            results = tqm.run(
                play_instance
            )  # most interesting data for a play is actually sent to the callback's methods
        finally:
            # we always need to cleanup child procs and the structures we use to communicate with
            tqm.cleanup()
            if loader:
                loader.cleanup_all_tmp_files()

        return results



"""

Metta provisioner plugin for testkit.

A provisioner plugin that uses testkit to create infra
and optionally to install either kubernetes or Mirantis
MKE/MSR onto the infra.

"""
import logging
import os
from typing import Any, Dict

import yaml

from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.jsonschema.validate import (
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL,
)
from configerus.validator import ValidationError

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures

from .client import TestkitClientPlugin, METTA_TESTKIT_CLIENT_PLUGIN_ID
from .testkit import TESTKITCLIENT_CLI_CONFIG_FILE_DEFAULT

logger = logging.getLogger("testkit.provisioner")

METTA_TESTKIT_PROVISIONER_PLUGIN_ID = "metta_testkit_provisioner"
""" Metta plugin id for testkit provisioner plugins """

TESTKIT_PROVISIONER_CONFIG_LABEL = "testkit"
"""Configerus config load label for accessing provisioner config."""
TESTKIT_PROVISIONER_CONFIG_BASE = LOADED_KEY_ROOT
""" configerus get base-key for accessing testkit provisioner config """

TESTKIT_CONFIG_KEY_SYSTEMNAME = "system_name"
""" config key to find the system name """
TESTKIT_CONFIG_KEY_SYSTEMS = "systems"
""" config key to find the set of systems installed by testkit """
TESTKIT_CONFIG_KEY_CREATE_OPTIONS = "options.create"
""" config key to find where the testkit create options """
TESTKIT_CONFIG_KEY_CONFIG = "config"
""" config key to find where to put the testkit config file data/contents """
TESTKIT_CONFIG_KEY_INSTANCES = "instances"
""" config key to find what instances to spec out (build) """
TESTKIT_CONFIG_KEY_CONFIGFILE = "config_file"
""" config key to find where to put the testkit config file """
TESTKIT_CONFIG_DEFAULT_CONFIGFILE = TESTKITCLIENT_CLI_CONFIG_FILE_DEFAULT
""" default value for where to put the testkit config file """

TESTKIT_CLIENT_SYSTEMS_KEY = "systems"
""" If provided, this config key provide a dictionary of configuration of client system plugins. """

TESTKIT_CONFIG_VALIDATE_JSONSCHEMA = {
    "type": "object",
    "properties": {
        "ucp": {"type": "object"},
        "dtr": {"type": "object"},
        "": {"type": "object"},
    },
    "required": [],
}
""" Validation jsonschema for testkit configuration for testkit yaml files """
TESTKIT_CONFIG_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: TESTKIT_CONFIG_VALIDATE_JSONSCHEMA
}
""" configerus jsonschema validation target for testkit config file """

TESTKIT_PROV_CONFIG_VAL_JSONSCHEMA = {
    "type": "object",
    "properties": {
        "opts": {"type": "object"},
        "config": TESTKIT_CONFIG_VALIDATE_JSONSCHEMA,
    },
    "required": [],
}
""" Validation jsonschema for provisioner configuration """
TESTKIT_PROV_CONFIG_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: TESTKIT_PROV_CONFIG_VAL_JSONSCHEMA
}
""" configerus jsonschema validation target for the provisioner plugin """


class TestkitProvisionerPlugin:
    """Testkit provisioner plugin.

    Provisioner plugin that allows control of and interaction with a testkit
    cluster.

    ## Requirements

    1. this plugin uses subprocess to call a testkit binary, so you have to
       install testkit in the environment

    ## Usage

    @TODO

    """

    def __init__(
        self,
        environment,
        instance_id,
        label: str = TESTKIT_PROVISIONER_CONFIG_LABEL,
        base: Any = TESTKIT_PROVISIONER_CONFIG_BASE,
    ):
        """Initialize Testkit provisioner.

        Parameters:
        -----------
        environment (Environment) : metta environment object that this plugin
            is attached.
        instance_id (str) : label for this plugin instances.

        label (str) : config load label for plugin configuration.
        base (str) : config base for loaded config for plugin configuration.

        """
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        self._config_label = label
        """ configerus load label that should contain all of the config """
        self._config_base = base

        self.fixtures = Fixtures()
        """This object makes and keeps track of fixtures for MKE/MSR clients."""

        try:
            self._write_config_file()
            self.make_fixtures()
            # pylint: disable= broad-except
        except Exception as err:
            # there are many reasons this can fail, and we just want to
            # see if we can get fixtures early.
            # No need to ask forgiveness for this one.
            logger.debug("Could not make initial fixtures: %s", err)

    # the deep argument is a standard for the info hook
    # pylint: disable=unused-argument
    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Get info about a provisioner plugin."""
        testkit_config = self._environment.config().load(self._config_label)

        return {
            "plugin": {
                "config": {
                    "config_label": self._config_label,
                    "config_base": self._config_base,
                },
                "config_file": testkit_config.get(
                    [self._config_base, TESTKIT_CONFIG_KEY_CONFIGFILE],
                    default="MISSING",
                ),
                "working_dir": testkit_config.get(
                    [self._config_base, TESTKIT_CONFIG_KEY_SYSTEMNAME], default="MISSING"
                ),
                "systems": testkit_config.get(
                    [self._config_base, TESTKIT_CONFIG_KEY_SYSTEMNAME],
                    default="MISSING",
                ),
            },
            "client": self._get_client_plugin().info(deep=deep),
        }

    def prepare(self):
        """Prepare any needed resources.

        We don't create the testkit file here so that it is created as late as
        possible.  This allows more options for dynamic config sources in the
        testkit config.

        """
        # Make sure that we are running on up to date config
        self._write_config_file()
        self.make_fixtures()

    def apply(self):
        """Create the testkit yaml file and run testkit to create a cluster."""
        # Make sure that we are running on up to date config
        self._write_config_file()
        self.make_fixtures()

        testkit_config = self._environment.config().load(self._config_label, force_reload=True)
        """ load the plugin configuration so we can retrieve options """
        opts = testkit_config.get(
            [self._config_base, TESTKIT_CONFIG_KEY_CREATE_OPTIONS], default={}
        )
        """ retrieve testkit client options from config """
        opt_list = []
        for key, value in opts.items():
            if isinstance(value, str):
                opt_list.append(f'--{key}="{value}"')
            else:
                opt_list.append(f"--{key}={value}")

        # run the testkit client command to provisioner the cluster
        self._get_client_plugin().create(opts=opt_list)

    def destroy(self):
        """Destroy any created resources."""
        # run the testkit client command to provisioner the cluster
        self._get_client_plugin().destroy()
        self._rm_config_file()

    def _write_config_file(self):
        """Write the config file for testkit."""
        try:
            # load all of the testkit configuration, force a reload to get up to date contents
            testkit_config = self._environment.config().load(self._config_label, force_reload=True)
            config = testkit_config.get(
                [self._config_base, TESTKIT_CONFIG_KEY_CONFIG],
                validator=TESTKIT_CONFIG_VALIDATE_TARGET,
            )
            """ config source of launchpad yaml """
        except KeyError as err:
            raise ValueError("Could not find launchpad configuration from config.") from err
        except ValidationError as err:
            raise ValueError("Launchpad config failed validation") from err

        config_file = testkit_config.get(
            [self._config_base, TESTKIT_CONFIG_KEY_CONFIGFILE],
            default=TESTKIT_CONFIG_DEFAULT_CONFIGFILE,
        )
        """ config_file value from plugin configuration """

        # write the configto our yaml file target (creating the path)
        os.makedirs(os.path.dirname(os.path.realpath(config_file)), exist_ok=True)
        with open(os.path.realpath(config_file), "w", encoding="utf8") as file:
            yaml.dump(config, file)

    def _rm_config_file(self):
        """Remove the written config file."""
        testkit_config = self._environment.config().load(self._config_label)
        config_file = testkit_config.get(
            [self._config_base, TESTKIT_CONFIG_KEY_CONFIGFILE],
            default=TESTKIT_CONFIG_DEFAULT_CONFIGFILE,
        )
        if os.path.isfile(config_file):
            os.remove(config_file)

    def make_fixtures(self):
        """Make related fixtures from a testkit installation.

        Creates:
        --------

        Testkit client : a client for interaction with the teskit cli

        """
        testkit_config = self._environment.config().load(self._config_label, force_reload=True)
        """ load the plugin configuration so we can retrieve options """

        try:
            testkit_config = self._environment.config().load(self._config_label, force_reload=True)
            """ loaded plugin configuration label """
        except KeyError as err:
            raise ValueError("Testkit plugin configuration did not have any config") from err

        system_name = testkit_config.get([self._config_base, TESTKIT_CONFIG_KEY_SYSTEMNAME])
        """ hat will testkit call the system """

        # instances = testkit_config.get([self._config_base, TESTKIT_CONFIG_KEY_INSTANCES])
        # """ what instances to create """

        config_file = testkit_config.get(
            [self._config_base, TESTKIT_CONFIG_KEY_CONFIGFILE],
            default=TESTKIT_CONFIG_DEFAULT_CONFIGFILE,
        )
        """ config_file value from plugin configuration """

        systems = testkit_config.get(
            [self._config_base, TESTKIT_CONFIG_KEY_SYSTEMS],
            default={},
        )

        fixture = self._environment.new_fixture(
            plugin_id=METTA_TESTKIT_CLIENT_PLUGIN_ID,
            instance_id=self.client_instance_id(),
            priority=70,
            arguments={
                "config_file": config_file,
                "system_name": system_name,
                "systems": systems,
            },
            labels={
                "parent_plugin_id": METTA_TESTKIT_PROVISIONER_PLUGIN_ID,
                "parent_instance_id": self._instance_id,
            },
            replace_existing=True,
        )
        # keep this fixture attached to the workload to make it retrievable.
        self.fixtures.add(fixture, replace_existing=True)

    def client_instance_id(self) -> str:
        """Construct an instanceid for the child client plugin."""
        return f"{self._instance_id}-{METTA_TESTKIT_CLIENT_PLUGIN_ID}"

    def _get_client_plugin(self) -> TestkitClientPlugin:
        """Retrieve the client plugin if we can."""
        try:
            return self.fixtures.get_plugin(instance_id=self.client_instance_id())
        except KeyError as err:
            raise RuntimeError(
                "Testkit provisioner cannot find its client plugin, and "
                "cannot process any client actions.  Was a client created?"
            ) from err



"""

Dynamic python module loading

"""
import os.path
import sys
import logging
import importlib
from importlib.util import spec_from_file_location, module_from_spec
from importlib.machinery import ModuleSpec
from types import ModuleType
from typing import List, Union, Any

from configerus.config import Config
from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.files import CONFIGERUS_PATH_KEY

logger = logging.getLogger("metta.import")

METTA_IMPORT_CONFIG_LABEL = "imports"
""" This config load() label is used to find modules to import """
METTA_IMPORT_CONFIG_IMPORTS_KEY = LOADED_KEY_ROOT
""" This config key is used to find modules to import """


def add_imports_from_config(
    config: Config,
    label: str = METTA_IMPORT_CONFIG_LABEL,
    base: Union[str, List[Any]] = METTA_IMPORT_CONFIG_IMPORTS_KEY,
):
    """Look in config for module imports.

    Use this if you want to dynamically import some modules defined in config.
    This can be used to load custom plugins that are decorated by the plugin
    Factory, but can include any module by path.

    @NOTE Modules are loaded by path, without a local namespace.

    Parameters:
    -----------
    config (Config) : config object to scan, and to add sources to

    label (str) : config label to load to search for sources
    base (str) : config key that should contain the list of sources

    """
    metta_config = config.load(label)

    imports_config = metta_config.get(base, default={})

    for import_name in imports_config:
        module_path = metta_config.get([base, import_name, CONFIGERUS_PATH_KEY])

        if os.path.isdir(module_path):
            module_path_dir = os.path.dirname(module_path)
            module_path_basename = os.path.basename(module_path)
            if not module_path_basename == import_name:
                logger.warning(
                    "Metta discovery importer cannot import a package (folder) using a"
                    "name other than the folder name: %s != %s",
                    module_path_basename,
                    import_name,
                )
            if module_path_dir not in sys.path:
                sys.path.append(module_path_dir)
            importlib.import_module(module_path_basename)
            logger.debug("Loaded package: %s : %s", module_path_basename, module_path)

        elif os.path.isfile(module_path):
            spec: ModuleSpec = spec_from_file_location(import_name, module_path)
            module: ModuleType = module_from_spec(spec)
            spec.loader.exec_module(module)
            logger.debug("Loaded module: %s : %s", import_name, module_path)

        else:
            raise ValueError(
                f"Could not import requested metta import {import_name} : {module_path}"
            )



"""

Global mutables.

Keep all package global variables in this module.

"""
from configerus.config import Config
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT as CONFIGERUS_SOURCE_DICT

from mirantis.testing.metta.fixture import Fixtures

# A global config object, which metta will create if needed.  This is not
# Absolutely required, but it makes sense in scenarios where Metta is asked
# to bootstrap itself and discover environments.
global_config: Config = Config()
global_config.add_source(
    plugin_id=CONFIGERUS_SOURCE_DICT, instance_id="metta-global", priority=10
).set_data({"metta": {"instance_id": "metta"}})

# The global fixtures set allows us to manage session/global based fixtures
# which allows us to keep components such as environments as plugins, and
# to use them in a global scope.
global_fixtures: Fixtures = Fixtures()
""" Keep a set fixtures which are kept in global scope """



"""

Ansible provisioner

"""

from typing import Any

from configerus.loaded import LOADED_KEY_ROOT

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.provisioner import METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER
from mirantis.testing.metta.client import METTA_PLUGIN_INTERFACE_ROLE_CLIENT
from mirantis.testing.metta.workload import METTA_PLUGIN_INTERFACE_ROLE_WORKLOAD
from mirantis.testing.metta_health.healthcheck import METTA_PLUGIN_INTERFACE_ROLE_HEALTHCHECK
from mirantis.testing.metta_cli.base import METTA_PLUGIN_INTERFACE_ROLE_CLI

from .ansiblecli_provisioner import (
    AnsiblePlaybookProvisionerPlugin,
    METTA_ANSIBLE_ANSIBLECLIPLAYBOOK_PROVISIONER_PLUGIN_ID,
    ANSIBLE_PROVISIONER_CONFIG_LABEL,
)
from .ansiblecli_client import (
    AnsibleClientPlugin,
    AnsiblePlaybookClientPlugin,
    METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID,
    METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKCLIENT_PLUGIN_ID,
)
from .ansiblecli_workload import (
    AnsibleCliCoreWorkloadPlugin,
    AnsibleCliPlaybookWorkloadPlugin,
    METTA_ANSIBLE_ANSIBLECLI_COREWORKLOAD_PLUGIN_ID,
    METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKWORKLOAD_PLUGIN_ID,
    ANSIBLE_WORKLOAD_CONFIG_LABEL,
)
from .cli import AnsibleCliPlugin, METTA_ANSIBLE_CLI_PLUGIN_ID


@Factory(
    plugin_id=METTA_ANSIBLE_ANSIBLECLIPLAYBOOK_PROVISIONER_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER],
)
def metta_plugin_factory_provisioner_ansibleplaybook(
    environment: Environment,
    instance_id: str = "",
    label: str = ANSIBLE_PROVISIONER_CONFIG_LABEL,
    base: Any = LOADED_KEY_ROOT,
):
    """create a metta provisioner plugin"""
    return AnsiblePlaybookProvisionerPlugin(environment, instance_id, label, base)


@Factory(
    plugin_id=METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLIENT, METTA_PLUGIN_INTERFACE_ROLE_HEALTHCHECK],
)
def metta_plugin_factory_client_ansibleplaybook(
    environment: Environment,
    instance_id: str = "",
    inventory_path: str = "",
    ansiblecfg_path: str = "",
):
    """create a metta client plugin"""
    return AnsibleClientPlugin(
        environment, instance_id, inventory_path=inventory_path, ansiblecfg_path=ansiblecfg_path
    )


@Factory(
    plugin_id=METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKCLIENT_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLIENT],
)
def metta_plugin_factory_client_ansible(
    environment: Environment,
    instance_id: str = "",
    inventory_path: str = "",
    ansiblecfg_path: str = "",
):
    """create a metta client plugin"""
    return AnsiblePlaybookClientPlugin(
        environment, instance_id, inventory_path=inventory_path, ansiblecfg_path=ansiblecfg_path
    )


@Factory(
    plugin_id=METTA_ANSIBLE_ANSIBLECLI_COREWORKLOAD_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_WORKLOAD],
)
def metta_plugin_factory_coreworkload_ansible(
    environment: Environment,
    instance_id: str = "",
    label: str = ANSIBLE_WORKLOAD_CONFIG_LABEL,
    base: Any = LOADED_KEY_ROOT,
):
    """create a metta workload plugin"""
    return AnsibleCliCoreWorkloadPlugin(environment, instance_id, label, base)


@Factory(
    plugin_id=METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKWORKLOAD_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_WORKLOAD],
)
def metta_plugin_factory_playbookworkload_ansible(
    environment: Environment,
    instance_id: str = "",
    label: str = ANSIBLE_WORKLOAD_CONFIG_LABEL,
    base: Any = LOADED_KEY_ROOT,
):
    """create a metta workload plugin"""
    return AnsibleCliPlaybookWorkloadPlugin(environment, instance_id, label, base)


@Factory(plugin_id=METTA_ANSIBLE_CLI_PLUGIN_ID, interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLI])
def metta_ansible_factory_cli_ansible(environment: Environment, instance_id: str = ""):
    """create a ansible cli plugin"""
    return AnsibleCliPlugin(environment, instance_id)


# ----- SetupTools EntryPoint METTA BootStrapping -----


# pylint: disable=unused-argument
def bootstrap_environment(environment: Environment):
    """METTA_Terraform bootstrap

    Currently we only use this to import plugins.

    Parameters:
    -----------

    env (Environment) : an environment which should have validation config added to.

    """



"""

Metta CLI : User plugin

Examine and declare user configuration for Metta

"""
import logging
import os

import appdirs
import yaml

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta_cli.base import CliBase, cli_output

from .common_config import METTA_COMMON_APP_NAME

logger = logging.getLogger("metta.cli.user")


METTA_PLUGIN_ID_CLI_USER = "user"
""" cli plugin_id for the user plugin """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class UserCliPlugin(CliBase):
    """Metta cli command generator for commands related to the current user."""

    def fire(self):
        """Return a dict of commands."""
        return {"user": UserGroup(self._environment)}


class UserGroup:
    """Commands relating to the current user.

    These commands can be used to manage the user/system settings for defining functionality that
    should apply to all projects, at the user level.

    """

    def __init__(self, environment: Environment):
        """Add environment to command group."""
        self._environment: Environment = environment

    # deep argument is an info() standard across plugins
    # pylint: disable=unused-argument
    def info(self, deep: bool = False):
        """Output any user related information."""
        user_config = self._environment.config().load("user")
        user_config_dir = _user_path()

        info = {
            "config": user_config.get(default={}),
            "system": {
                "path": user_config_dir,
                "exists": os.path.isdir(user_config_dir),
            },
        }

        return cli_output(info)

    # pylint: disable=no-self-use
    def init(self):
        """Make a User config source that will be global for this system."""
        user_config_dir = _user_path()
        if _user_path_exists():
            raise RuntimeError(
                f"User has already been initialized on this system: {user_config_dir}"
            )

        os.mkdir(user_config_dir)

    # pylint: disable=no-self-use
    def set_id(self, uid: str):
        """Set the user id for the user on the system."""
        if not _user_path_exists():
            raise RuntimeError(
                "No user configuration has been initialized on this system.  Run `init` first"
            )

        user_config_dir = _user_path()

        config_path_user = os.path.join(user_config_dir, "user.yml")
        try:
            with open(config_path_user, "r", encoding="utf8") as user_config_file:
                user_config = yaml.load(user_config_file)
        except FileNotFoundError:
            user_config = {}

        user_config["id"] = uid

        with open(config_path_user, "w", encoding="utf8") as user_config_file:
            yaml.dump(user_config, user_config_file)


def _user_path():
    """Get path to user specific configuration."""
    # a user config path (like ~/.config/metta) may contain config
    return appdirs.user_config_dir(METTA_COMMON_APP_NAME)


def _user_path_exists():
    """Perform boolean check on whether or not the user config path exists on the system."""
    # a user config path (like ~/.config/metta) may contain config
    user_config_dir = _user_path()
    return os.path.isdir(user_config_dir)



"""

ansible cli handler.

Here are is all of the functionality which actually runs ansible commands
on the command line using subprocess.

"""

import logging
import json
import os
import subprocess
import shutil
from typing import Dict, List, Any

logger = logging.getLogger("metta_ansible:ansible-cli")

ANSIBLE_CLIENT_DEFAULT_BINARY = "ansible"
""" default ansible executable for subprocess """
ANSIBLEPLAYBOOK_CLIENT_DEFAULT_BINARY = "ansible-playbook"
""" default ansible-playbook executable for subprocess """

METTA_ANSIBLE_ANSIBLECLI_ENV_JSONOUT = {
    "ANSIBLE_STDOUT_CALLBACK": "json",
    "ANSIBLE_LOAD_CALLBACK_PLUGINS": "1",
}
"""Passing these env vars to ansible will give us json output."""


class AnsibleClient:
    """Shell client for running ansible-playbook using subprocess."""

    # this is what it takes to configure the terraform client
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        inventory_path: str,
        ansiblecfg_path: str = "",
        ansiblebinary: str = ANSIBLE_CLIENT_DEFAULT_BINARY,
    ):
        """Initialize Ansible client.

        Parameters:
        -----------
        inventory_path (str) : string path to the ansible inventory file
        ansiblecfg_path (str) : options string path to an ansiblecfg file

        """
        self.inventory_path = inventory_path
        """String path to an ansible invetory path."""
        self.ansiblecfg_path = ansiblecfg_path
        """Optional string path to an ansible cfg file."""

        if shutil.which(ansiblebinary) is None:
            raise ValueError(
                "Ansible binary not found. Ansible commands cannot be called. "
                f"Expected binary at path {ansiblebinary}"
            )

        self.ansible_bin = ansiblebinary
        """Ansible binary executable path to run."""

    # deep argument is a part of the info() interface for all plugins
    # pylint: disable=unused-argument
    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Return dict of plugin info for introspection."""
        return {
            "inventory_path": self.inventory_path,
            "ansiblecfg_path": self.ansiblecfg_path,
        }

    def debug(self, hosts: str = "all"):
        """Run ansible debug and return parsed response."""
        return json.loads(
            self.run(
                ["--module-name=debug", hosts],
                envs=METTA_ANSIBLE_ANSIBLECLI_ENV_JSONOUT,
                return_output=True,
            )
        )

    def setup(self, hosts: str = "all"):
        """Run ansible setup and return parsed response."""
        return json.loads(
            self.run(
                ["--module-name=setup", hosts],
                envs=METTA_ANSIBLE_ANSIBLECLI_ENV_JSONOUT,
                return_output=True,
            )
        )

    def ping(self, hosts: str = "all"):
        """Run ansible ping and return parsed response."""
        return json.loads(
            self.run(
                ["--module-name=ping", hosts],
                envs=METTA_ANSIBLE_ANSIBLECLI_ENV_JSONOUT,
                return_output=True,
            )
        )

    def run(
        self,
        args: List[str],
        envs: Dict[str, str] = None,
        with_inventory=True,
        with_ansiblecfg=True,
        return_output=False,
    ):
        """Run ansible CLI command."""
        env = os.environ.copy()

        cmd = [self.ansible_bin]

        if with_ansiblecfg and self.ansiblecfg_path:
            env["ANSIBLE_CONFIG"] = self.ansiblecfg_path
        if with_inventory:
            cmd += [f"--inventory={self.inventory_path}"]
        if envs is not None:
            env.update(envs)

        cmd += args

        # improve readability
        # pylint: disable=no-else-return
        if not return_output:
            logger.debug("running ansible command: %s", " ".join(cmd))
            res = subprocess.run(cmd, env=env, check=True, text=True)
            res.check_returncode()
            return res
        else:
            logger.debug("running ansible command with output capture: %s", " ".join(cmd))
            return_res = subprocess.run(
                cmd, env=env, shell=False, check=True, stdout=subprocess.PIPE
            )
            return_res.check_returncode()
            return return_res.stdout.decode("utf-8")


class AnsiblePlaybookClient:
    """Shell client for running ansible-playbook using subprocess."""

    # this is what it takes to configure the terraform client
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        inventory_path: str,
        ansiblecfg_path: str,
        playbookbinary: str = ANSIBLEPLAYBOOK_CLIENT_DEFAULT_BINARY,
    ):
        """Initialize Ansible Playbook client.

        Parameters:
        -----------
        inventory_path (str) : string path to the ansible inventory file
        ansiblecfg_path (str) : options string path to an ansiblecfg file
        """
        self.inventory_path = inventory_path
        """String path to an ansible invetory path."""
        self.ansiblecfg_path = ansiblecfg_path
        """Optional string path to an ansible cfg file."""

        if shutil.which(playbookbinary) is None:
            raise ValueError(
                "Ansible-Playbook binary not found. Ansible commands cannot be called. "
                f"Expected binary at path {playbookbinary}"
            )

        self.ansibleplaybook_bin = playbookbinary
        """Ansible-Playbook binary executable path to run."""

    # deep argument is a part of the info() interface for all plugins
    # pylint: disable=unused-argument
    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Return dict of plugin info for introspection."""
        return {
            "inventory_path": self.inventory_path,
            "ansiblecfg_path": self.ansiblecfg_path,
        }

    def run(
        self,
        playbooksyml_path: str,
        extravars_path: str = "",
        extravars: Dict[str, str] = None,
        envs: Dict[str, str] = None,
        with_inventory=True,
        with_ansiblecfg=True,
        return_output=False,
    ):
        """Run the ansible playbook install command on a yaml playbook file."""
        args = [playbooksyml_path]

        return self._run(
            args=args,
            extravars_path=extravars_path,
            extravars=extravars,
            envs=envs,
            return_output=return_output,
        )

    def _run(
        self,
        args: List[str],
        extravars_path: str = "",
        extravars: Dict[str, str] = None,
        envs: Dict[str, str] = None,
        with_inventory=True,
        with_ansiblecfg=True,
        return_output=False,
    ):
        """Run ansible-playbook CLI command."""
        allenvs = os.environ.copy()

        cmd = [self.ansibleplaybook_bin]

        if with_ansiblecfg and self.ansiblecfg_path:
            allenvs["ANSIBLE_CONFIG"] = self.ansiblecfg_path
        if extravars_path:
            cmd += [f"--extra-vars=@{extravars_path}"]
        if extravars:
            extravars_string = json.dumps(extravars)
            cmd += [f"--extra-vars='{extravars_string}'"]
        if with_inventory:
            cmd += [f"--inventory={self.inventory_path}"]
        if envs is not None:
            allenvs.update(envs)

        cmd += args

        # improve readability
        # pylint: disable=no-else-return
        if not return_output:
            logger.debug("running ansible-playbook command: %s", " ".join(cmd))
            res = subprocess.run(cmd, env=allenvs, check=True, text=True)
            res.check_returncode()
            return res
        else:
            logger.debug("running ansible-playbook command with output capture: %s", " ".join(cmd))
            return_res = subprocess.run(
                cmd, env=allenvs, shell=False, check=True, stdout=subprocess.PIPE
            )
            return_res.check_returncode()
            return return_res.stdout.decode("utf-8")



"""

METTA Kubernetes.

metta contrib functionality for Kubernetes.  In particular to provide plugins
for kubernetes clients and kubernetes workloads.

"""

from typing import Any

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.client import METTA_PLUGIN_INTERFACE_ROLE_CLIENT
from mirantis.testing.metta_health.healthcheck import METTA_PLUGIN_INTERFACE_ROLE_HEALTHCHECK
from mirantis.testing.metta.workload import METTA_PLUGIN_INTERFACE_ROLE_WORKLOAD
from mirantis.testing.metta_cli.base import METTA_PLUGIN_INTERFACE_ROLE_CLI

from .kubeapi_client import (
    KubernetesApiClientPlugin,
    METTA_PLUGIN_ID_KUBERNETES_CLIENT,
)
from .deployment_workload import (
    KubernetesDeploymentWorkloadPlugin,
    METTA_PLUGIN_ID_KUBERNETES_DEPLOYMENT_WORKLOAD,
    KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_LABEL,
    KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_BASE,
)
from .yaml_workload import (
    KubernetesYamlWorkloadPlugin,
    METTA_PLUGIN_ID_KUBERNETES_YAML_WORKLOAD,
    KUBERNETES_YAML_WORKLOAD_CONFIG_LABEL,
    KUBERNETES_YAML_WORKLOAD_CONFIG_BASE,
)
from .helm_workload import (
    KubernetesHelmWorkloadPlugin,
    METTA_PLUGIN_ID_KUBERNETES_HELM_WORKLOAD,
    KUBERNETES_HELM_WORKLOAD_CONFIG_LABEL,
    KUBERNETES_HELM_WORKLOAD_CONFIG_BASE,
)
from .cli import KubernetesCliPlugin, METTA_PLUGIN_ID_KUBERNETES_CLI


@Factory(
    plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT,
    interfaces=[
        METTA_PLUGIN_INTERFACE_ROLE_CLIENT,
        METTA_PLUGIN_INTERFACE_ROLE_HEALTHCHECK,
        METTA_PLUGIN_ID_KUBERNETES_CLIENT,
    ],
)
def metta_plugin_factory_client_kubernetes(
    environment: Environment, instance_id: str = "", kube_config_file: str = ""
) -> KubernetesApiClientPlugin:
    """Create a metta kubernetes client plugin."""
    return KubernetesApiClientPlugin(environment, instance_id, kube_config_file)


@Factory(
    plugin_id=METTA_PLUGIN_ID_KUBERNETES_DEPLOYMENT_WORKLOAD,
    interfaces=[
        METTA_PLUGIN_INTERFACE_ROLE_WORKLOAD,
        METTA_PLUGIN_INTERFACE_ROLE_HEALTHCHECK,
    ],
)
def metta_plugin_factory_workload_kubernetes_deployment(
    environment: Environment,
    instance_id: str = "",
    label: str = KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_LABEL,
    base: Any = KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_BASE,
):
    """Create a metta kubernetes spec workload plugin."""
    return KubernetesDeploymentWorkloadPlugin(environment, instance_id, label=label, base=base)


@Factory(
    plugin_id=METTA_PLUGIN_ID_KUBERNETES_YAML_WORKLOAD,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_WORKLOAD],
)
def metta_plugin_factory_workload_kubernetes_yaml(
    environment: Environment,
    instance_id: str = "",
    label: str = KUBERNETES_YAML_WORKLOAD_CONFIG_LABEL,
    base: Any = KUBERNETES_YAML_WORKLOAD_CONFIG_BASE,
):
    """Create a metta kubernetes spec workload plugin."""
    return KubernetesYamlWorkloadPlugin(environment, instance_id, label=label, base=base)


@Factory(
    plugin_id=METTA_PLUGIN_ID_KUBERNETES_HELM_WORKLOAD,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_WORKLOAD, METTA_PLUGIN_INTERFACE_ROLE_HEALTHCHECK],
)
def metta_plugin_factory_workload_kubernetes_helm(
    environment: Environment,
    instance_id: str = "",
    label: str = KUBERNETES_HELM_WORKLOAD_CONFIG_LABEL,
    base: Any = KUBERNETES_HELM_WORKLOAD_CONFIG_BASE,
):
    """Create a metta kubernetes spec workload plugin."""
    return KubernetesHelmWorkloadPlugin(environment, instance_id, label=label, base=base)


@Factory(
    plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLI,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLI],
)
def metta_terraform_factory_cli_kubernetes(environment: Environment, instance_id: str = ""):
    """Create a kubernetes cli plugin."""
    return KubernetesCliPlugin(environment, instance_id)


# ----- SetupTools EntryPoint METTA BootStrapping -----


# pylint: disable=unused-argument
def bootstrap_environment(environment: Environment):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment which should have validation config added to.

    """



"""

Fixture management code.

A Fixture is a plugin instance wrapper that keeps metadata about the plugin
instance.
A Fixtures object is a collection of plugin instances kept as an
ordered, managed set.

A fixture saves us from applying any required interface onto a plugin for
maintaining metadata about the plugin, however, as a wrapper, it also leaves
the plugin in the dark about its own metadata.

## TODO

* Fixtures are now a pivotal cental component in metta, because most of the
  core and contrib functionality is provided by plugins, even bootstrapping
  and environments are plugins.  This means that even the core uses a lot of
  Fixtures searching and managemeng.
  Because of this prevalence, the Fixtures object could use some streamlining
  and optimization for its implementation.
  Some ideas:
    1. low volume (<100) optimization of filtering/searching methods; perhaps
       caching.
    2. optimization of merging and copying?

"""
import logging
from typing import Any, Dict, List, Iterator

# pylint: disable=W0511
# TODO move these to this file as METTA_FIXTURE_KEY_XXXXX
from .plugin import (
    Instance,
    METTA_PLUGIN_CONFIG_KEY_PLUGINID,
    METTA_PLUGIN_CONFIG_KEY_INSTANCEID,
)

logger = logging.getLogger("metta.fixture")

METTA_FIXTURES_CONFIG_FIXTURES_LABEL = "fixtures"
""" A centralized configerus load label for multiple fixtures """
METTA_FIXTURES_CONFIG_FIXTURE_KEY = "fixture"
""" Config .get() key for a single fixture """
METTA_FIXTURE_CONFIG_KEY_PRIORITY = "priority"
""" configerus .get()  assign an instance a priority when it is created. """
METTA_FIXTURE_CONFIG_KEY_CONFIG = "config"
""" configerus .get()  as additional config """
METTA_FIXTURE_CONFIG_KEY_VALIDATORS = "validators"
""" configerus .get()  to decide what validators to apply to the plugin """

METTA_FIXTURE_VALIDATION_JSONSCHEMA = {
    "type": "object",
    "properties": {
        METTA_PLUGIN_CONFIG_KEY_PLUGINID: {"type": "string"},
        METTA_PLUGIN_CONFIG_KEY_INSTANCEID: {"type": "string"},
        METTA_FIXTURE_CONFIG_KEY_PRIORITY: {
            "type": "integer",
            "minimum": 1,
            "maximum": 100,
        },
    },
    "required": [METTA_PLUGIN_CONFIG_KEY_PLUGINID],
}
""" json schema validation definition for a plugin """


class Fixture:
    """A plugin wrapper struct that keep metadata about the plugin in a set.

    A Fixture is yet another wrapper for a plugin object which contains
    metadata about the plugin.  This is similar to the plugin.PluginInstance
    wrapper, but it also adds a priority intereger for relative importance.

    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        plugin: Any,
        plugin_id: str,
        instance_id: str,
        interfaces: List[str],
        labels: Dict[str, str],
        priority: int,
    ):
        """Initialize struct contents.

        Parameters:
        -----------
        plugin : the fixture plugin instance

        Filtering parameters:

        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier
        interfaces (list[str]) : string list of interface
            identifiers that the plugin should support.

        """
        self.plugin_id: str = plugin_id
        self.instance_id: str = instance_id
        self.interfaces: List[str] = interfaces
        self.labels: Dict[str, str] = labels
        self.priority: int = priority
        self.plugin: Any = plugin

    def __eq__(self, other):
        """Compare to another fixture.

        Parameters:
        -----------
        other (Fixture) : fixture to compare this fixture against.

        Returns:
        --------
        Boolean True if the other fixture matches this one in metadata,
        otherwise False.

        """
        if not isinstance(other, Fixture):
            return False

        return self.plugin_id == other.plugin_id and self.instance_id == other.instance_id

    def info(self, deep: bool = False, children: bool = False) -> Dict[str, Any]:
        """Return some dict metadata about the fixture and plugin.

        Parameters:
        -----------
        deep (bool) : whether or not the information provided should go deep
            into the avaiable sources of data.  Info is intended to be cheap
            to perform, but there are some scenarios where you want more data
            despite the cost, or risk for exception.
        children (bool) : if the fixture has children fixtures, then this
            flag controls whether or not the children info should also be
            included.

        Returns:
        --------
        Dict[str, Any] or primitives that can be used for introspection of the
        fixtures.

        Raises:
        -------
        Info() should be safe for operation, but if deep=True then there may be
        a risk of a plugin throwing an exception.

        """
        fixture_info: Dict[str, Any] = {
            "fixture": {
                "instance_id": self.instance_id,
                "plugin_id": self.plugin_id,
                "interfaces": self.interfaces,
                "labels": self.labels,
                "priority": self.priority,
            }
        }
        # If a plugin has a callable info method then add it to the info
        if hasattr(self.plugin, "info"):
            fixture_info["plugin"] = self.plugin.info(deep=deep)
        # if the plugin hsa child, optionally add their info
        if children and hasattr(self.plugin, "fixtures") and callable(self.plugin.fixtures):
            fixture_info["fixtures"] = {}
            for fixture in self.plugin.fixtures():
                fixture_info["fixtures"][fixture.instance_id] = fixture.info(
                    deep=deep, children=children
                )

        return fixture_info

    def has_interfaces(self, interfaces: List[str]) -> bool:
        """Does this fixture have all of the passed interfaces."""
        for required_interface in interfaces:
            if required_interface not in self.interfaces:
                return False
        return True

    def has_labels(self, labels: List[str]) -> bool:
        """Does this fixture have all of the passed labels."""
        for required_label in labels:
            if required_label not in self.labels.keys():
                return False
        return True

    @classmethod
    def from_instance(
        cls, instance: Instance, priority: int, labels: Dict[str, str] = None
    ) -> "Fixture":
        """Convert a plugin instance to a fixture."""
        # Labels are merged from the plugin instance and the passed in fixture overrides.
        if labels is None:
            labels = {}
        if instance.labels is not None:
            labels_copy = instance.labels.copy()
            labels_copy.update(labels)
            labels = labels_copy

        return cls(
            plugin_id=instance.plugin_id,
            instance_id=instance.instance_id,
            interfaces=instance.interfaces,
            labels=labels,
            priority=priority,
            plugin=instance.plugin,
        )


class Fixtures:
    """A managed set of Fixture objects.

    A set of plugins that can be added in an arbitrary order but retrieved
    using filters and sorting.  The plugins are wrapped as Fixtures, which are
    used for the sorting and filtering.

    To use this, create an instance, and then add fixtures using either new()
    or add().  You can also combine two collections using merge().
    To get fixtures out, either iterator across the Fixtures object, or use
    get() to retrieve a single matching item.  use filter() to produce a new
    Fixtures object as a reduced set.

    """

    def __init__(self):
        """Create initial empty fixtures list."""
        self._fixtures: List[Fixture] = []
        """ object List of fixtures. """

    def __len__(self) -> int:
        """Return how many plugin instances we have.

        Returns:
        --------
        Integer length of the collection of Fixture objects.

        """
        return len(self._fixtures)

    def __getitem__(self, instance_id: str):
        """Subscribe to a fixtures instance using instance_id.

        Parameters:
        -----------
        instance_id (str) : fixture instance_id you are looking for.

        Returns:
        --------
        A Fixture object with the passed instance_id.

        Raises:
        -------
        Will raise a KeyError if the instance_id does not exist in the set. the
        exception is actually raised by the .get() method.

        """
        return self.get(instance_id=instance_id)

    def __iter__(self) -> Iterator[Fixture]:
        """Create an iterator for the fixtures object.

        @TODO switch to just using a generator?

        Returns:
        --------
        An Iterator of Fixture objects

        """
        # Iterate across the to_list() set, as it is sorted.
        return iter(self.to_list())

    def __reversed__(self) -> Iterator[Fixture]:
        """Create a reversed iterator for the fixtures object.

        @TODO switch to just using a generator?

        Returns:
        --------
        An Iterator of Fixture objects

        """
        # Iterate across the to_list() set, as it is sorted.
        return reversed(self.to_list())

    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Return some dict metadata about the fixtures and plugins."""
        fixtures_info: Dict[str, Any] = {}
        for fixture in self._fixtures:
            fixtures_info[fixture.instance_id] = fixture.info(deep=deep)

        return fixtures_info

    def merge(self, merge_from: "Fixtures"):
        """Merge fixture instances from another Fixtures object into this one.

        Parameters:
        -----------
        Merge_from (Fixtures) : fixture instance source

        Raises:
        -------
        May raise a KeyError if there a matching plugin is already in the
        Fixtures object.

        """
        # We use the add_fixture method to centralize the logic for adding fixtures to one function
        for fixture in merge_from:
            self.add(fixture)

    # pylint: disable=too-many-arguments
    def new(
        self,
        plugin: object,
        plugin_id: str,
        instance_id: str,
        interfaces: List[str],
        labels: Dict[str, str],
        priority: int,
        replace_existing: bool = False,
    ) -> Fixture:
        """Add a new fixture by providing the plugin instance and the metadata.

        Create a new Fixture from the passed arguments and add it to the
        Fixtures set.

        Parameters:
        -----------
        plugin : the fixture plugin instance

        Filtering parameters:

        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier
        interfaces (List[str]) : interface identifiers supported
        labels: [str, str] : labels for the instance
        priority (int) : plugin priority

        """
        return self.add(
            Fixture(
                plugin_id=plugin_id,
                instance_id=instance_id,
                priority=priority,
                interfaces=interfaces,
                labels=labels,
                plugin=plugin,
            ),
            replace_existing=replace_existing,
        )

    def add(self, fixture: Fixture, replace_existing: bool = False) -> Fixture:
        """Add an existing fixture.

        If a matching

        Parameters:
        -----------
        fixture (Fixture) : existing fixture to add

        replace_existing (bool) : If True, then this fixture can replace a
            matching plugin with the same metadata.

        """
        if self.get(
            plugin_id=fixture.plugin_id,
            instance_id=fixture.instance_id,
            exception_if_missing=False,
        ):

            if not replace_existing:
                raise KeyError(
                    "Fixture index already exists:"
                    f"[plugin_id:{fixture.plugin_id}]"
                    f"[instance_id:{fixture.instance_id}]"
                )

            # array index needed for replacement
            # pylint: disable=consider-using-enumerate
            for i in range(len(self._fixtures)):
                if self._fixtures[i] == fixture:
                    self._fixtures[i] = fixture
                    return fixture

        self._fixtures.append(fixture)

        return fixture

    def get(
        self,
        plugin_id: str = "",
        instance_id: str = "",
        interfaces: List[str] = None,
        labels: Dict[str, str] = None,
        has_labels: List[str] = None,
        exception_if_missing: bool = True,
    ) -> Fixture:
        """Retrieve the first matching fixture object based on filters and priority.

        Parameters:
        -----------
        Filtering parameters:

        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier
        interfaces (List[str]) : List of interfaces which the fixture must have
        labels (Dict[str, str]) : Filter key value pairs which fixtures must match
        has_labels (List[str]) : List of labels which the fixture must have

        Returns:
        --------
        The highest priority matched Fixture fixture.
        If now fixtures matched, and exception_if_missing is False, then None

        Raises:
        -------
        KeyError if exception_if_missing is True and no matching fixture was
        found.

        """
        filtered = self.filter(
            plugin_id=plugin_id,
            instance_id=instance_id,
            interfaces=interfaces,
            labels=labels,
            has_labels=has_labels,
        )

        if len(filtered) > 0:
            return filtered.to_list()[0]
        if exception_if_missing:
            raise KeyError(
                "Could not find any matching fixture instances "
                f"[plugin_id:{plugin_id}][instance_id:{instance_id}]"
                f"[interfaces:{interfaces}][labels:{labels}]"
            )
        # filtered list was empty, and we were not directed to raise an exception for that.
        return None

    def get_plugin(
        self,
        plugin_id: str = "",
        instance_id: str = "",
        interfaces: List[str] = None,
        labels: Dict[str, str] = None,
        has_labels: List[str] = None,
        exception_if_missing: bool = True,
    ) -> object:
        """Retrieve the first matching plugin  based on filters and priority.

        Parameters:
        -----------
        Filtering parameters:

        plugin_id (str) : registry plugin_id
        interfaces (List[str]) : List of interfaces which the fixture must have
        labels (Dict[str, str]) : Filter key value pairs which fixtures must match
        has_labels (List[str]) : List of labels which the fixture must have
        instance_id (str) : plugin instance identifier

        Returns:
        --------
        The highest priority matched Fixture fixture.
        If now fixtures matched, and exception_if_missing is False, then None

        Raises:
        -------
        KeyError if exception_if_missing is True and no matching fixture was
        found. This error is actually raise in self.get()

        """
        fixture = self.get(
            plugin_id=plugin_id,
            instance_id=instance_id,
            interfaces=interfaces,
            labels=labels,
            has_labels=has_labels,
            exception_if_missing=exception_if_missing,
        )

        if fixture is None:
            return None
        return fixture.plugin

    # I should try to fix the branch count here
    # pylint: disable=too-many-branches
    def filter(
        self,
        plugin_id: str = "",
        instance_id: str = "",
        interfaces: List[str] = None,
        labels: Dict[str, str] = None,
        has_labels: List[str] = None,
        exception_if_missing: bool = False,
    ) -> "Fixtures":
        """Filter the fixture instances.

        Parameters:
        -----------
        Filtering parameters:

        plugin_id (str) : registry plugin_id
        interfaces (List[str]) : List of interfaces which the fixture must have
        labels (Dict[str, str]) : Filter key value pairs which fixtures must match
        has_labels (List[str]) : List of labels which the fixture must have
        instance_id (str) : plugin instance identifier

        Returns:
        --------
        A new Fixtures object with only matching Fixture objects.  It could
        contain all of the items. If no filters were passed in, or it could
        be empty if no matches were found and the passed exception_is_missing
        variable is True

        Raises:
        -------
        KeyError if exception_if_missing is True and no matching fixture was found

        """
        filtered = Fixtures()
        for fixture in self._fixtures:
            try:
                if interfaces:
                    for required_interface in interfaces:
                        if required_interface not in fixture.interfaces:
                            raise DoesNotMatchError("does not contain required interface")

                if has_labels:
                    for required_label in has_labels:
                        if required_label not in fixture.labels:
                            raise DoesNotMatchError("does not contain required label")

                if labels:
                    for key, value in labels.items():
                        if key not in fixture.labels:
                            raise DoesNotMatchError("does not contain required label")
                        if not fixture.labels[key] == value:
                            raise DoesNotMatchError("required label does not match")

                if plugin_id and not fixture.plugin_id == plugin_id:
                    raise DoesNotMatchError("plugin_id does not match")
                if instance_id and not fixture.instance_id == instance_id:
                    raise DoesNotMatchError("instance_id does not match")

                filtered.add(fixture, replace_existing=True)

            except DoesNotMatchError:
                pass

        if exception_if_missing and len(filtered) == 0:
            raise KeyError(f"Filter found matches [{plugin_id}][{instance_id}]")
        return filtered

    def filter_out(
        self,
        plugin_id: str = "",
        instance_id: str = "",
        interfaces: List[str] = None,
        labels: Dict[str, str] = None,
        has_labels: List[str] = None,
        exception_if_missing: bool = False,
    ) -> "Fixtures":
        """Return the fixtures with fixtures removed.

        Parameters:
        -----------
        Filtering parameters:

        plugin_id (str) : registry plugin_id
        interfaces (List[str]) : List of interfaces which the fixture must have
        labels (List[str]) : List of labels which the fixture must have
        instance_id (str) : plugin instance identifier

        Returns:
        --------
        A new Fixtures object with only matching Fixture objects.  It could
        contain all of the items. If no filters were passed in, or it could
        be empty if no matches were found and the passed exception_is_missing
        variable is True

        Raises:
        -------
        KeyError if exception_if_missing is True and no matching fixture was found

        """
        filtered = Fixtures()
        for fixture in self._fixtures:
            try:
                if interfaces:
                    for required_interface in interfaces:
                        if required_interface not in fixture.interfaces:
                            raise DoesNotMatchError("does not contain required interface")

                if has_labels:
                    for required_label in has_labels:
                        if required_label not in fixture.labels:
                            raise DoesNotMatchError("does not contain required label")

                if labels:
                    for key, value in labels.items():
                        if key not in fixture.labels:
                            raise DoesNotMatchError("does not contain required label")
                        if not fixture.labels[key] == value:
                            raise DoesNotMatchError("required label does not match")

                if plugin_id and not fixture.plugin_id == plugin_id:
                    raise DoesNotMatchError("plugin_id does not match")

                if instance_id and not fixture.instance_id == instance_id:
                    raise DoesNotMatchError("instance_id does not match")

            except DoesNotMatchError:
                filtered.add(fixture, replace_existing=True)

        if exception_if_missing and len(filtered) == 0:
            raise KeyError(f"Filter found matches [{plugin_id}][{instance_id}]")
        return filtered

    def to_list(self) -> List[Fixture]:
        """Order the fixtures based on priority.

        This is usefull if you need a list, but you can rely on the
        Fixtures.__iter__ for most scenarios, and just iterate across the
        Fixtures object itself.

        This method is relied on internally.

        Returns:
        --------
        A List[Fixture] of the contained fixtures sorted using their priority,
        sorted from lowest to highest priority.

        """
        return sorted(self._fixtures, key=lambda i: 1 / i.priority if i.priority > 0 else 100)


class DoesNotMatchError(Exception):
    """An exception to indicate that a value does not match filter expectations."""



"""

Metta cli plugin for testkit.

Mainly provides functionality for interacting with the testkit provisioner.

"""
import logging

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta_cli.base import CliBase, cli_output

from .provisioner import METTA_TESTKIT_PROVISIONER_PLUGIN_ID
from .client import METTA_TESTKIT_CLIENT_PLUGIN_ID

logger = logging.getLogger("metta.cli.testkit")

METTA_TESTKIT_CLI_PLUGIN_ID = "metta_testkit_cli"
""" metta plugin id for the testkit cli plugin """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class TestkitCliPlugin(CliBase):
    """Fire command/group generator for testkit commands."""

    def fire(self):
        """Return a dict of commands."""
        commands = {}

        if (
            self._environment.fixtures().get(
                plugin_id=METTA_TESTKIT_CLIENT_PLUGIN_ID,
                exception_if_missing=False,
            )
            is not None
        ):
            commands["testkit"] = TestkitClientGroup(self._environment)

        return commands


class TestkitClientGroup:
    """Testkit client commands."""

    def __init__(self, environment: Environment):
        """Inject environment."""
        self._environment: Environment = environment

        if (
            self._environment.fixtures().get(
                plugin_id=METTA_TESTKIT_PROVISIONER_PLUGIN_ID,
                exception_if_missing=False,
            )
            is not None
        ):
            self.provisioner = TestkitProvisionerGroup(self._environment)

    def _select_client(self, instance_id: str = ""):
        """Pick a matching client."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=METTA_TESTKIT_CLIENT_PLUGIN_ID,
                instance_id=instance_id,
            )

        # Get the highest priority provisioner
        return self._environment.fixtures().get(
            plugin_id=METTA_TESTKIT_CLIENT_PLUGIN_ID,
        )

    def info(self, client: str = "", deep: bool = False):
        """Get info about a client plugin."""
        fixture = self._select_client(instance_id=client)
        return cli_output(fixture.info(deep=deep))

    def create(self, client: str = ""):
        """Run  testkit create."""
        client_plugin = self._select_client(instance_id=client).plugin
        return cli_output(client_plugin.create())

    def destroy(self, client: str = ""):
        """Run testkit system rm."""
        client_plugin = self._select_client(instance_id=client).plugin
        return cli_output(client_plugin.destroy())

    def systems(self, client: str = ""):
        """List the systems that testkit is aware of."""
        client_plugin = self._select_client(instance_id=client).plugin
        return cli_output(client_plugin.system_ls())

    def machines(self, client: str = ""):
        """List the systems that testkit is aware of."""
        client_plugin = self._select_client(instance_id=client).plugin
        return cli_output(client_plugin.machine_ls())

    def exec(self, host: str, cmd: str, client: str = ""):
        """List the systems that testkit is aware of."""
        client_plugin = self._select_client(instance_id=client).plugin
        client_plugin.exec(host=host, cmd=cmd)


class TestkitProvisionerGroup:
    """Base Fire command group for testkit provisioner commands."""

    def __init__(self, environment: Environment):
        """Inject environment."""
        self._environment: Environment = environment

    def _select_provisioner(self, instance_id: str = ""):
        """Pick a matching provisioner."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=METTA_TESTKIT_PROVISIONER_PLUGIN_ID,
                instance_id=instance_id,
            )

        # Get the highest priority provisioner
        return self._environment.fixtures().get(
            plugin_id=METTA_TESTKIT_PROVISIONER_PLUGIN_ID,
        )

    def info(self, provisioner: str = "", deep: bool = False):
        """Get info about a provisioner plugin."""
        fixture = self._select_provisioner(instance_id=provisioner)
        return cli_output(fixture.info(deep=deep))

    def prepare(self, provisioner: str = ""):
        """Run the provisioner prepare : which for testkit currently does nothing."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        return cli_output(provisioner_plugin.prepare())

    def apply(self, provisioner: str = ""):
        """Run the provisioner apply, which runs testkit create."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        return cli_output(provisioner_plugin.apply())

    def destroy(self, provisioner: str = ""):
        """Run the provisioner destroy, which runs testkit system rm."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        return cli_output(provisioner_plugin.destroy())
